# **************************************************************************
# *
# * Authors:     Pablo Conesa (pconesa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import platform
import abc
from abc import ABC

from pyworkflow.gui.project.constants import STATUS_COLORS, WARNING_COLOR
from pyworkflow.protocol import STATUS_FAILED


def getStatusColorFromNode(node):
    # If it is a run node (not PROJECT)
    return getStatusColorFromRun(node.run)


def getStatusColorFromRun(prot):
    """ Returns the color associated with the status. """
    if prot:
        if prot.hasSummaryWarnings():
            return WARNING_COLOR
        else:
            return getStatusColor(prot.status.get(STATUS_FAILED))
    else:
        return getStatusColor()


def getStatusColor(status=None, default='#ADD8E6'):
    """
    Parameters
    ----------
    status status of the protocol

    Returns the color associated with he status
    -------

    """
    return STATUS_COLORS[status] if status else default

# OS dependent behaviour. Add any OS dependent method here and later we might move
# or refactor this to a class or something else


class OSHandler(abc.ABC):
    """ Abstract class: Handler for OS specific actions"""
    def maximizeWindow(root):
        pass


class LinuxHandler(OSHandler, ABC):

    def maximizeWindow(root):
        root.attributes("-zoomed", True)


class MacHandler(OSHandler, ABC):

    def maximizeWindow(root):
        root.state("zoomed")


class WindowsHandler(OSHandler, ABC):

    def maximizeWindow(root):
        root.state("zoomed")


class OS:
    _handler = None

    _handlers = {"Linux": LinuxHandler,
                 "Darwin": MacHandler,
                 "Windows": WindowsHandler}  # Until testing this on windows

    @staticmethod
    def getPlatform():
        return platform.system()

    @classmethod
    def handler(cls):
        if cls._handler is None:
            cls._handler = cls._handlers[cls.getPlatform()]

        return cls._handler





# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABIlITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This modules holds the base classes for the ORM implementation.
The Object class is the root in the hierarchy and some other
basic classes.
"""
import os
from collections import OrderedDict
import datetime as dt

from pyworkflow import utils
from pyworkflow.utils.reflection import getSubclasses


# Binary relations always involve two objects, we 
# call them parent-child objects, the following
# constants reflect which direction of the relation we refer
RELATION_CHILDS = 0
RELATION_PARENTS = 1

# There are other relations related to data provenance,
# which object were used to produce a result (source of it)
RELATION_SOURCE = 'relation_datasource'
RELATION_TRANSFORM = 'relation_transform'


# Then column name of the parent id relation
OBJECT_PARENT_ID = 'object_parent_id'


class Object(object):
    """ All objects in our Domain should inherit from this class
    that will contains all base properties"""

    def __init__(self, value=None, **kwargs):
        object.__init__(self)
        self._objIsPointer = kwargs.get('objIsPointer', False)  # True if will be treated as a reference for storage
        self._objId = kwargs.get('objId', None)  # Unique identifier of this object in some context
        self._objParentId = kwargs.get('objParentId', None)  # identifier of the parent object
        self._objName = kwargs.get('objName', '')  # The name of the object will contains the whole path of ancestors
        self._objLabel = kwargs.get('objLabel', '')  # This will serve to label the objects
        self._objComment = kwargs.get('objComment', '')
        self._objTag = kwargs.get('objTag', None)  # This attribute serve to make some annotation on the object.
        self._objDoStore = kwargs.get('objDoStore', True)  # True if this object will be stored from his parent
        self._objCreation = None
        self._objParent = None  # Reference to parent object
        self._objEnabled = True
        self.set(value)

    def getClassName(self):
        return self.__class__.__name__
    
    def getClass(self):
        return type(self)
    
    @classmethod
    def getDoc(cls):
        return cls.__doc__ or ''

    # FIXME: This function should be renamed to hasAttribute when we address that issue
    def hasAttributeExt(self, attrName):
        attrList = attrName.split('.')
        obj = self
        for partName in attrList:
            obj = getattr(obj, partName, None)
            if obj is None:
                return False
        return True

    # FIXME: This function is not symmetric with setAttributeValue
    def hasAttribute(self, attrName):
        return hasattr(self, attrName)

    # FIXME: This function is not symmetric with setAttributeValue
    def getAttributeValue(self, attrName, defaultValue=None):
        """ Get the attribute value given its name.
        Equivalent to getattr(self, name).get() 
        """
        attr = getattr(self, attrName, None)
        if attr is None:
            value = defaultValue
        elif callable(attr):
            value = attr()
        elif isinstance(attr, Object):
            value = attr.get()
        else:
            value = attr  # behave well for non-Object attributes
        return value
    
    def setAttributeValue(self, attrName, value, ignoreMissing=False):
        """ Set the attribute value given its name.
        Equivalent to setattr(self, name).set(value) 
        If the attrName contains dot: x.y
        it will be equivalent to getattr(getattr(self, 'x'), 'y').set(value)
        If ignoreMissing is True, unexisting attrName will not raise an
        exception.
        """
        attrList = attrName.split('.')
        obj = self
        for partName in attrList:
            obj = getattr(obj, partName, None)
            if obj is None:
                if ignoreMissing:
                    return
                raise Exception("Object.setAttributeValue: obj is None! attrName: "
                                "%s, part: %s" % (attrName, partName))
        obj.set(value)
        
    def getAttributes(self):
        """Return the list of attributes than are
        subclasses of Object"""
        for key, attr in self.__dict__.items():
            if issubclass(attr.__class__, Object):
                yield key, attr
                
    def getAttributesToStore(self):
        """Return the list of attributes than are
        subclasses of Object and will be stored"""
        for key, attr in self.getAttributes():
            if not hasattr(attr, '_objDoStore'):
                print("Object.getAttributesToStore: attribute '%s' seems to "
                      "be overwritten," % key)
                print("   since '_objDoStore' was not found. "
                      "Ignoring attribute. ")
            else:
                if attr is not None and attr._objDoStore:
                    yield key, attr

    def isPointer(self):
        """If this is true, the value field is a pointer 
        to another object"""
        return self._objIsPointer
        
    def _convertValue(self, value):
        """Convert a value to desired scalar type"""
        return value
    
    def set(self, value):
        """Set the internal value, if it is different from None
        call the convert function in subclasses"""
        if value is not None:
            value = self._convertValue(value)            
        self._objValue = value
    
    def get(self):
        """Return internal value"""
        return self._objValue
    
    def trace(self, callback):
        """ Add an observer when the set method is called. """
        if self.set == self.__setTrace:
            pass
        else:
            self.__set = self.set 
            self.set = self.__setTrace
        self.__setCallback = callback 
        
    def __setTrace(self, value):
        self.__set(value)
        self.__setCallback()
    
    def getObjValue(self):
        """Return the internal value for storage.
        This is a good place to do some update of the
        internal value before been stored"""
        return self._objValue
    
    def getObjId(self):
        """Return object id"""
        return self._objId
    
    def setObjId(self, newId):
        """Set the object id"""
        self._objId = newId
        
    def copyObjId(self, other):
        """ Copy the object id form other to self. """
        self.setObjId(other.getObjId())
        
    def hasObjId(self):
        return self._objId is not None
    
    def cleanObjId(self):
        """ This function will set to None this object id
        and the id of all its children attributes.
        This function should be used when retrieving
        an object from a mapper and storing in a different one.
        """
        self.setObjId(None)
        for _, attr in self.getAttributesToStore():
            attr.cleanObjId()
            
    def getObjParentId(self):
        return self._objParentId
    
    def hasObjParentId(self):
        return self._objParentId is not None
            
    def getObjLabel(self):
        """ Return the label associated with this object"""
        return self._objLabel
    
    def setObjLabel(self, label):
        """ Set the label to better identify this object"""
        self._objLabel = label
             
    def getObjComment(self):
        """ Return the comment associated with this object"""
        return self._objComment
    
    def setObjComment(self, comment):
        """ Set the comment to better identify this object"""
        self._objComment = comment       
        
    def setObjCreation(self, creation):
        """ Set the creation time of the object. """
        self._objCreation = creation
        
    def getObjCreation(self):
        """ Return the stored creation time of the object. """
        return self._objCreation

    def getObjCreationAsDate(self):
        """ Return the stored creation time of the object as date """
        return String.getDatetime(self._objCreation)

    def strId(self):
        """String representation of id"""
        return str(self._objId)
    
    def getName(self):
        # TODO: REMOVE THIS FUNCTION, SINCE IT DOES NOT COMPLAIN WITH _objX naming
        return self._objName
    
    def getObjName(self):
        return self._objName
    
    def setEnabled(self, enabled):
        self._objEnabled = bool(enabled)
        
    def isEnabled(self):
        """Return if object is enabled"""
        return self._objEnabled
    
    def getNameId(self):
        """ Return an unique and readable id that identifies this object. """
        label = self.getObjLabel()
        if len(label) > 0:
            return label
        elif self.hasObjId():
            return '%s.%s' % (self.getName(), self.strId())
        return ''
    
    def getLastName(self):
        """ If the name contains parent path, remove it
        and only return the attribute name in its parent. 
        """
        if '.' in self._objName:
            return self._objName.split('.')[-1]
        return self._objName 
    
    def setName(self, name):
        self._objName = name
        
    def hasValue(self):        
        return True
    
    def getStore(self):
        """Return True if the object will be stored by the mapper"""
        return self._objDoStore
    
    def setStore(self, value):
        """set the store flag"""
        self._objDoStore = value

    def __hash__(self):
        # Issue found in scipion-em-xmipp/xmipp3/protocols/protocol_extract_particles_pairs.py (_setupBasicProperties)
        # TypeError: unhashable type: 'Micrograph'
        # Solution: Hashability makes an object usable as a dictionary key and a set member, because these data
        # structures use the hash value internally. All of Python’s immutable built_in objects are hashable, while no
        # mutable containers (such as lists or dictionaries) are. Objects which are instances of user-defined classes
        # are hashable by default; they all compare unequal, and their hash value is their id().
        # https://docs.python.org/3.1/glossary.html
        return id(self)

    def __eq__(self, other):
        """Comparison for scalars should be by value
        and for other objects by reference"""
        if self._objValue is None:
            return object.__eq__(self, other)
        return self._objValue == other._objValue
    
    def equalAttributes(self, other, ignore=[], verbose=False):
        """Compare that all attributes are equal"""
        for k, v1 in self.getAttributes():
            # v1 = getattr(self, k) # This is necessary because of FakedObject simulation of getattr
            # Skip comparison of attribute names in 'ignore' list
            if k in ignore:
                continue
            v2 = getattr(other, k)
            if issubclass(type(v1), Object):
                comp = v1.equalAttributes(v2, ignore=ignore, verbose=verbose)
            else:
                comp = v1 == v2
            if not comp:
                if verbose:
                    print("Different attributes: ")
                    print("self.%s = %s" % (k, v1))
                    print("other.%s = %s" % (k, v2))
                return False
        return True
            
    def copyAttributes(self, other, *attrNames):
        """ Copy attributes in attrNames from other to self. 
        If the name X is in attrNames, it would be equivalent to:
        self.X.set(other.X.get())
        This method is more useful for Scalar attributes.
        There are two patchs for Pointer and PointerList.
        """
        for name in attrNames:
            attr = getattr(self, name, None)
            otherAttr = getattr(other, name)

            if attr is None:
                setattr(self, name, otherAttr.clone())
            elif isinstance(attr, Pointer):
                attr.copy(otherAttr)
            elif isinstance(attr, PointerList):
                for pointer in otherAttr:
                    attr.append(pointer)
            else:
                attr.set(otherAttr.get())
            
    def __getObjDict(self, prefix, objDict, includeClass):
        if prefix:
            prefix += '.'
        for k, v in self.getAttributesToStore():
            if not v.isPointer():
                kPrefix = prefix + k
                if includeClass:
                    objDict[kPrefix] = (v.getClassName(), v.getObjValue())
                else:
                    objDict[kPrefix] = v.getObjValue()
                if not isinstance(v, Scalar):
                    v.__getObjDict(kPrefix, objDict, includeClass)
            
    def getObjDict(self, includeClass=False, includeBasic=False):
        """ Return all attributes and values in a dictionary.
        Nested attributes will be separated with a dot in the dict key.
        Params:
            includeClass: if True, the values will be a tuple (ClassName, value)
                otherwise only the values of the attributes
            includeBasic: if True include the id, label and comment.
                object.id: objId
                object.label: objLabel
                object.comment: objComment
        """
        d = OrderedDict()

        if includeClass:
            d['self'] = (self.getClassName(),)

        if includeBasic:
            d['object.id'] = self.getObjId()
            d['object.label'] = self.getObjLabel()
            d['object.comment'] = self.getObjComment()

        self.__getObjDict('', d, includeClass)

        return d

    def setAttributesFromDict(self, attrDict, setBasic=True,
                              ignoreMissing=False):
        """ Set object attributes from the dict obtained from getObjDict.
         WARNING: this function is yet experimental and not fully tested.
        """
        if setBasic:
            self.setObjId(attrDict.get('object.id', None))
            self.setObjLabel(attrDict.get('object.label', ''))
            self.setObjComment(attrDict.get('object.comment', ''))

        for attrName, value in attrDict.items():
            if not attrName.startswith('object.'):
                self.setAttributeValue(attrName, value, ignoreMissing)
            
    def __getMappedDict(self, prefix, objDict):
        if prefix:
            prefix += '.'
        for k, v in self.getAttributesToStore():
            if not v.isPointer():
                kPrefix = prefix + k
                objDict[kPrefix] = v
                if not isinstance(v, Scalar):
                    v.__getMappedDict(kPrefix, objDict)
                        
    def getMappedDict(self):
        d = OrderedDict()
        self.__getMappedDict('', d)
        return d        
    
    def getNestedValue(self, key):
        """ Retrieve the value of nested attributes like: _ctfModel.defocusU. """
        attr = self
        for p in key.split('.'):
            attr = getattr(attr, p)
        return attr.get()
    
    def getValuesFromDict(self, objDict):
        """ Retrieve the values of the attributes
        for each of the keys that comes in the objDict.
        """
        return [self.getNestedValue(k) for k in objDict if k != 'self']
    
    def getValuesFromMappedDict(self, mappedDict):
        return [v.getObjValue() for v in mappedDict.values()]
    
    def copy(self, other, copyId=True, ignoreAttrs=[]):
        """ Copy all attributes values from one object to the other.
        The attributes will be created if needed with the corresponding type.
        Params:
            other: the other object from which to make the copy.
            copyId: if true, the _objId will be also copied.
            ignoreAttrs: pass a list with attributes names to ignore.
        """
        copyDict = {'internalPointers': []} 
        self._copy(other, copyDict, copyId, ignoreAttrs=ignoreAttrs)
        self._updatePointers(copyDict)
        return copyDict
        
    def _updatePointers(self, copyDict):
        """ Update the internal pointers after a copy. 
        If there are pointers to other object in the copy 
        the references should be updated.
        """
        for ptr in copyDict['internalPointers']:
            pointedId = ptr.getObjValue().getObjId()
            if pointedId in copyDict:
                ptr.set(copyDict[pointedId])
        
    def _copy(self, other, copyDict, copyId, level=1, ignoreAttrs=[]):
        """ Recursively clone all attributes from one object to the other.
        (Currently, we are not deleting attributes missing in the 'other' object.)
        Params:
        copyDict: this dict is used to store the ids map between 'other' and
            'self' attributes. It is used for update pointers and relations
            later on. This will only work if the ids of 'other' attributes
            has been properly set.
        """
        # Copy basic object data
        # self._objName = other._objName
        if copyId:
            self._objId = other._objId
        self._objValue = other._objValue
        self._objLabel = other._objLabel
        self._objComment = other._objComment
        # Copy attributes recursively
        for name, attr in other.getAttributes():
            if name not in ignoreAttrs:
                myAttr = getattr(self, name, None)
    
                if myAttr is None:
                    myAttr = attr.getClass()()
                    setattr(self, name, myAttr)
                    
                myAttr._copy(attr, copyDict, copyId, level+2)
                # Store the attr in the copyDict
                if attr.hasObjId():
                    # storing in copyDict with id=", attr.getObjId()
                    copyDict[attr.getObjId()] = myAttr
                # Use the copyDict to fix the reference in the copying object
                # if the pointed one is inside the same object
                if myAttr.isPointer() and myAttr.hasValue():
                    copyDict['internalPointers'].append(myAttr)
    
    def clone(self):
        clone = self.getClass()()
        clone.copy(self)        
        return clone    
    
    def evalCondition(self, condition):
        """ Check if condition is meet.
        Params:
            condition: the condition string, it can contains variables
                or methods without arguments to be evaluated.
            Examples:
                hasCTF
                hasCTF and not hasAlignment
        Return:
            The value of the condition evaluated with values
        """
        # Split in possible tokens
        import re
        tokens = re.split('\W+', condition)
        condStr = condition
        
        for t in tokens:
            if self.hasAttribute(t):
                condStr = condStr.replace(t, str(self.getAttributeValue(t)))
        return eval(condStr)
    
    def printAll(self, name=None, level=0):
        """Print object and all its attributes.
        Mainly for debugging"""
        tab = ' ' * (level*3)
        idStr = ''  # ' (id = %s, pid = %s)' % (self.getObjId(), self._objParentId)
        if name is None:
            print(tab, self.getClassName(), idStr)
        else:
            if name == 'submitTemplate':  # Skip this because very large value
                value = '...'
            else:
                value = self.getObjValue()
                
            print(tab, '%s = %s' % (name, value), idStr)
        for k, v in self.getAttributes():
            v.printAll(k, level + 1)
            
    def printObjDict(self, includeClasses=False):
        """Print object dictionary. Mainly for debugging"""
        import pprint
        pp = pprint.PrettyPrinter(indent=4)
        pp.pprint(dict(self.getObjDict(includeClasses)))        


class OrderedObject(Object):
    """This is based on Object, but keep the list
    of the attributes to store in the same order
    of insertion, this can be useful where order matters"""
    def __init__(self, value=None, **kwargs):
        object.__setattr__(self, '_attributes', [])
        Object.__init__(self, value, **kwargs)

    def __attrPointed(self, name, value):
        """ Check if a value is already pointed by other
        attribute. This will prevent to storing pointed
        attributes such as:
        self.inputMics = self.inputMicrographs.get()
        In this case we want to avoid to store self.inputMics as 
        another attribute of this object.
        """
        for key in self._attributes:
            attr = getattr(self, key)
            if attr is not None and attr.isPointer():
                if attr.get() is value:
                    return True
        return False
    
    def __setattr__(self, name, value):
        if (name not in self._attributes and
            issubclass(value.__class__, Object) and
                not self.__attrPointed(name, value) and value._objDoStore):
            self._attributes.append(name)
        Object.__setattr__(self, name, value)
    
    def getAttributes(self):
        """Return the list of attributes than are
        subclasses of Object and will be stored"""
        for key in self._attributes:
            yield key, getattr(self, key)
                
    def deleteAttribute(self, attrName):
        """ Delete an attribute. """
        if attrName in self._attributes:
            self._attributes.remove(attrName)
            delattr(self, attrName)

                
class Scalar(Object):
    """Base class for basic types"""
    def hasValue(self):        
        return self._objValue is not None
    
    def equalAttributes(self, other, ignore=[], verbose=False):
        """Compare that all attributes are equal"""
        return self._objValue == other._objValue
    
    def __str__(self):
        """String representation of the scalar value"""
        return str(self.get())

    def __cmp(self, other):
        """ Comparison implementation for scalars. """
        a = self.get()
        b = other.get() if isinstance(other, Object) else other
        # cmp does not longer exist in Python3, so we will follow
        # the recommend replacement: (a > b) - (a < b)
        # https://docs.python.org/3.0/whatsnew/3.0.html#ordering-comparisons

        # Consider None values
        if a is None and b is None:
            return True
        elif a is None or b is None:
            return False
        else:
            return (a > b) - (a < b)

    def __eq__(self, other):
        """ Comparison for scalars should be by value while
         for other objects by reference. """
        return self.__cmp(other) == 0

    def __ne__(self, other):
        return self.__cmp(other) != 0

    def __lt__(self, other):
        return self.__cmp(other) < 0

    def __le__(self, other):
        return self.__cmp(other) <= 0

    def __gt__(self, other):
        return self.__cmp(other) > 0

    def __ge__(self, other):
        return self.__cmp(other) >= 0

    def get(self, default=None):
        """Get the value, if internal value is None
        the default argument passed is returned. """
        if self.hasPointer():
            return self._pointer.get().get(default)

        if self.hasValue():
            return self._objValue
        return default
    
    def _copy(self, other, *args, **kwargs):
        self.set(other.get())
        
    def swap(self, other):
        """ Swap the contained value between
        self and other objects.
        """
        tmp = self._objValue
        self._objValue = other._objValue
        other._objValue = tmp

    def sum(self, value):
        self._objValue += self._convertValue(value)
        
    def multiply(self, value):
        self._objValue *= value

    def setPointer(self, pointer):
        """ Set an internal pointer from this Scalar.
        Then, the value (retrieved with get) will be obtained
        from the pointed object.
        """
        if pointer is None:
            if self.hasPointer():
                delattr(self, "_pointer")
        else:
            self._pointer = pointer

    def hasPointer(self):
        """ Return True if this Scalar has an internal pointer
        from where the value will be retrieved with the get() method.
        """
        return hasattr(self, '_pointer')

    def getPointer(self):
        """ Return the internal pointer of this Scalar or None. """
        return getattr(self, '_pointer', None)

    
class Integer(Scalar):
    """Integer object"""
    def _convertValue(self, value):
        return int(value)
    
    def increment(self):
        """ Add 1 to the current value. """
        self._objValue += 1
        
    def __float__(self):
        return float(self.get())
    
    def __int__(self):
        return self.get()
    
    def __long__(self):
        return int(self.get())
    
        
class String(Scalar):
    """String object. """
    DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"
    FS = ".%f"  # Femto seconds

    @classmethod
    def getDatetime(cls, strValue, formatStr=None, fs=True):
        """ Get the datetime from the given string value.
        Params:
            strValue: string representation of the date
            formatStr: if is None, use the default DATETIME_FORMAT.
            fs: Use femto seconds or not, only when format=None
        """
        if formatStr is None:
            try:
                formatStr = cls.DATETIME_FORMAT
                if fs:
                    formatStr += cls.FS
                datetime = dt.datetime.strptime(strValue, formatStr)
            except Exception as ex:
                # Maybe the %f (femtoseconds) is not working
                # let's try to format without it
                datetime = dt.datetime.strptime(strValue, cls.DATETIME_FORMAT)
        else:
            datetime = dt.datetime.strptime(strValue, formatStr)

        return datetime

    def _convertValue(self, value):
        return str(value)
    
    def empty(self):
        """ Return true if None or len == 0 """
        if not self.hasValue():
            return True
        return len(self.get().strip()) == 0

    def datetime(self, formatStr=None, fs=True):
        """ Get the datetime from this object string value. """
        return String.getDatetime(self._objValue, formatStr, fs)


class Float(Scalar):
    """Float object"""
    EQUAL_PRECISION = 0.001
    
    @classmethod
    def setPrecision(cls, newPrecision):
        """ Set the precision to compare float values.
        Mainly used for testing purposes.
        """
        cls.EQUAL_PRECISION = newPrecision
        
    def _convertValue(self, value):
        return float(value)
    
    def equalAttributes(self, other, ignore=[], verbose=False):
        """Compare that all attributes are equal"""
        # If both float has some value distinct of None
        # then we should compare the absolute value of difference 
        # against the EQUAL_PRECISION to say equal or not
        if self.hasValue() and other.hasValue():
            return abs(self._objValue - other._objValue) < self.EQUAL_PRECISION
        # If one has None as value, then both should have None
        # to have equal attributes
        if not self.hasValue() and not other.hasValue():
            return True
        
        return False
    
    def __float__(self):
        return self.get()
        
        
class Boolean(Scalar):
    """Boolean object"""
    
    def _convertValue(self, value):
        t = type(value)
        if t is bool:
            return value
        if t is str:
            v = value.strip().lower()
            return v == 'true' or v == '1'
        return bool(value) 
    
    def __nonzero__(self):
        if not self.hasValue():
            return False
        return self.get() 
    
    def __bool__(self):
        return self.get()  
    
    
class Pointer(Object):
    """Reference object to other one"""
    EXTENDED_ATTR = '__attribute__'
    EXTENDED_ITEMID = '__itemid__'
    _ERRORS = {}
    
    def __init__(self, value=None, **kwargs):
        Object.__init__(self, value, objIsPointer=True, **kwargs)
        # The _extended attribute will be used to point to attributes of a
        # pointed object or the id of an item inside a set
        self._extended = String()
        
        if 'extended' in kwargs:
            self.setExtended(kwargs.get('extended')) 
               
    def __str__(self):
        """String representation of a pointer"""
        if self.hasValue():
            className = self.getObjValue().getClassName()
            strId = self.getObjValue().strId()
            return '-> %s (%s)' % (className, strId)
        return '-> None'
    
    def __clean(self, extended):
        """ Remove old attributes conventions. """
        # TODO: This replacements are needed now by backward compatibility
        # reasons, when we used __attribute__ and __item__ to mark both cases
        # in a future the following two lines can be removed.
        ext = extended.replace(self.EXTENDED_ATTR, '')
        ext = ext.replace(self.EXTENDED_ITEMID, '')
        return ext
        
    def hasValue(self):
        return self._objValue is not None
    
    def get(self, default=None):
        """ Get the pointed object. 
        By default all pointers store a "pointed object" value.
        The _extended attribute allows to also point to internal
        attributes or items (in case of sets) of the pointed object.
        """
        extended = self._extended.get()
        if extended:
            ext = self.__clean(extended)
            parts = ext.split('.')
            value = self._objValue
            for p in parts:
                if p.isdigit():
                    value = value[int(p)]  # item case
                else:
                    value = getattr(value, p, None)
                if value is None:
                    break
        else:
            value = self._objValue
            
        return value
    
    def set(self, other):
        """ Set the pointer value but cleaning the extended property. """
        Object.set(self, other)
        # This check is needed because set is call from the Object constructor
        # when this attribute is not setup yet (a dirty patch, I know)
        if hasattr(self, '_extended'):
            self._extended.set(None)
        
    def hasExtended(self):
        return bool(self._extended.get())  # consider empty string as false
    
    def getExtended(self):
        return self.__clean(self._extended.get(''))
        
    def setExtended(self, attribute):
        """ Set the attribute name of the "pointed object"
        that will be the result of the get() action. 
        """
        self._extended.set(attribute)
    
    def getExtendedParts(self):
        """ Return the extended components as a list. """
        if self.hasExtended():
            return self.getExtended().split('.')
        else:
            return []
    
    def setExtendedParts(self, parts):
        """ Set the extedend attribute but using 
        a list as input. 
        """
        self.setExtended('.'.join(parts))
        
    def addExtended(self, attribute):
        """ Similar to setExtended, but concatenating more extensions
        instead of replacing the previous value.
        """
        if self.hasExtended():
            self._extended.set('%s.%s' % (self._extended.get(), attribute))
        else:
            self.setExtended(attribute)
            
    def removeExtended(self):
        """ Remove the last part of the extended attribute. """
        if self.hasExtended():
            parts = self.getExtendedParts()
            self.setExtendedParts(parts[:-1])
        
    def getAttributes(self):
        yield '_extended', getattr(self, '_extended')
        
    def pointsNone(self):
        return self.get() is None
    
    def getUniqueId(self):
        """ Return an unique id concatenating the id
        of the direct pointed object plus the extended 
        attribute.
        """
        uniqueId = self.getObjValue().strId()
        if self.hasExtended():
            uniqueId += '.%s' % self.getExtended()
        
        return uniqueId
    

class List(Object, list):
    ITEM_PREFIX = '__item__'
    
    """Class to store a list of objects"""
    def __init__(self, value=None, **kwargs):
        list.__init__(self)
        Object.__init__(self, value, **kwargs)
        
    def __getattr__(self, name):
        if name.startswith(self.ITEM_PREFIX):
            i = self._stringToIndex(name)
            if i < len(self):
                return self[i]
        raise AttributeError("List object has not attribute: " + name)
            
    def __setattr__(self, name, value):
        if name.startswith('__item__') or len(name) == 0:
            self.append(value)
        else:
            object.__setattr__(self, name, value)

    def getAttributes(self):
        # First yield all attributes not contained in the list
        for name, attr in Object.getAttributes(self):
            yield name, attr
        # Now yield elements contained in the list
        for i, item in enumerate(self):
            yield self._indexToString(i), item
            
    def _indexToString(self, i):
        """Return the way the string index is generated.
        String indexes will start in 1, that's why i+1
        """
        return "%s%06d" % (self.ITEM_PREFIX, i+1)
    
    def _stringToIndex(self, strIndex):
        """ From the string index representation obtain the index.
        For symetry the number in the index string will be
        decreased in 1.
        """
        return int(strIndex.split(self.ITEM_PREFIX)[1]) - 1
            
    def __len__(self):
        return list.__len__(self)

    def getSize(self):  # Just to have similar API than Set
        return len(self)
    
    def isEmpty(self):
        return len(self) == 0
    
    def clear(self):
        del self[:]
        
    def _convertValue(self, value):
        """Value should be a list."""
        if not (isinstance(value,  list)):
            raise Exception("List.set: value should be a list.")
        self.clear()
        for item in value:
            self.append(item)
        return None
    
        
class PointerList(List):
    def __init__(self, value=None, **kwargs):
        List.__init__(self, value, **kwargs)
        
    def append(self, value):
        """ Append Pointer of objects to the list.
         If value is a Pointer, just add it to the list.
         If is another subclass of Object, create
         a Pointer first and append the pointer.
        """
        if isinstance(value, Pointer):
            pointer = value
        elif isinstance(value, Object):
            pointer = Pointer()
            pointer.set(value)
        else:
            raise Exception("Only subclasses of Object can be added to PointerList\n"
                            " Passing value: %s, type: %s" % (value, type(value)))

        List.append(self, pointer)

            
class CsvList(Scalar, list):
    """This class will store a list of objects
    in a single DB row separated by comma.
    pType: the type of the list elements, int, bool, str"""
    def __init__(self, pType=str, **kwargs):
        Scalar.__init__(self, **kwargs)
        list.__init__(self)
        self._pType = pType
        
    def _convertValue(self, value):
        """ Value should be a str with comma separated values or a list.
        """
        self.clear()
        if value:
            if isinstance(value, str):
                for s in value.split(','):
                    self.append(self._pType(s))
            elif isinstance(value, list) or isinstance(value, tuple):
                for s in value:
                    self.append(self._pType(s))
            else:
                raise Exception("CsvList.set: Invalid value type: ",
                                type(value))
            
    def getObjValue(self):
        self._objValue = ','.join(map(str, self))
        return self._objValue
    
    def get(self):
        return self.getObjValue()
    
    def __str__(self):
        return list.__str__(self)
     
    def isEmpty(self):
        return len(self) == 0
    
    def clear(self):
        del self[:]

    def __eq__(self, other):
        """ Comparison for scalars should be by value
        and for other objects by reference.
        """
        return all(a == b for a, b in zip(self, other))


class Set(OrderedObject):
    """ This class will be a container implementation for elements.
    It will use an extra sqlite file to store the elements.
    All items will have an unique id that identifies each element in the set.
    """
    ITEM_TYPE = None  # This property should be defined to know the item type
    
    # This will be used for stream Set where data is populated on the fly
    STREAM_OPEN = 1
    STREAM_CLOSED = 2

    FILE_TEMPLATE_NAME = 'set%s.sqlite'
    indexes = ['_index']
    
    def __init__(self, filename=None, prefix='', 
                 mapperClass=None, classesDict=None, **kwargs):
        # Use the object value to store the filename
        OrderedObject.__init__(self, **kwargs)
        self._mapper = None
        self._idCount = 0
        self._size = Integer(0)  # cached value of the number of images
        # It is a bit contradictory that initially a set is Closed
        # but this is the default behaviour of the Set before Streamming extension
        self._streamState = Integer(self.STREAM_CLOSED)  
        self.setMapperClass(mapperClass)
        self._mapperPath = CsvList()  # sqlite filename
        self._representative = None
        self._classesDict = classesDict
        self._indexes = kwargs.get('indexes', [])

        # If filename is passed in the constructor, it means that
        # we want to create a new object, so we need to delete it if
        # the file exists
        if filename:
            self._mapperPath.set('%s, %s' % (filename, prefix))
            self.load()
            
    def _getMapper(self):
        """ This method will open the connection to the items
        database on demand. That's why this method should 
        be used instead of accessing directly _mapper.
        """
        if self._mapper is None:
            self.load()
        return self._mapper
            
    def aggregate(self, operations, operationLabel, groupByLabels=None):
        return self._getMapper().aggregate(operations, operationLabel, groupByLabels)

    def setMapperClass(self, MapperClass):
        """ Set the mapper to be used for storage. """
        if MapperClass is None:
            from pyworkflow.mapper.sqlite import SqliteFlatMapper
            MapperClass = SqliteFlatMapper
        Object.__setattr__(self, '_MapperClass', MapperClass)
        
    def __getitem__(self, itemId):
        """ Get the image with the given id. """
        return self._getMapper().selectById(itemId)

    def __contains__(self, itemId):
        """ element in Set """
        return self._getMapper().exists(itemId)

    def iterItems(self, orderBy='id', direction='ASC', where='1',
                  limit=None):
        return self._getMapper().selectAll(orderBy=orderBy,
                                           direction=direction,
                                           where=where,
                                           limit=limit)  # has flat mapper, iterate is true

    def getFirstItem(self):
        """ Return the first item in the Set. """
        # This function is used in many contexts where the mapper can be
        # left open and could be problematic locking other db
        # So, we looking if the mapper was closed before, in which case
        # we will close it after that
        closedMapper = self._mapper is None
        firstItem = self._getMapper().selectFirst()
        if closedMapper:
            self.close()
        return firstItem
    
    def __iter__(self):
        """ Iterate over the set of images. """
        return self.iterItems()
       
    def __len__(self):
        return self._size.get()
    
    def getSize(self):
        """Return the number of images"""
        return self._size.get()

    def isEmpty(self):
        return self.getSize() == 0
    
    def getFileName(self):
        if len(self._mapperPath):
            return self._mapperPath[0]
        return None
    
    def getPrefix(self):
        if len(self._mapperPath) > 1:
            return self._mapperPath[1]
        return None
    
    def write(self, properties=True):
        """
        Commit the changes made to the Set underlying database.
        Params:
            properties: this flag controls when to write Set attributes to 
                special table 'Properties' in the database. False value is 
                use for example in SetOfClasses for not writing each Class2D 
                properties.
        """
        if properties:
            self._getMapper().setProperty('self', self.getClassName())
            objDict = self.getObjDict()
            for key, value in objDict.items():
                self._getMapper().setProperty(key, value)
        self._getMapper().commit()
    
    def _loadClassesDict(self):
        return self._classesDict or globals()
    
    def setClassesDict(self, classesDict):
        """ Set the dictionary with classes where to look for classes names. """
        self._classesDict = classesDict
    
    def load(self):
        """ Load extra data from files. """
        if self._mapperPath.isEmpty():
            raise Exception("Set.load:  mapper path and prefix not set.")
        fn, prefix = self._mapperPath
        self._mapper = self._MapperClass(fn, self._loadClassesDict(), prefix, self._indexes)
        self._size.set(self._mapper.count())
        self._idCount = self._mapper.maxId()
           
    def __del__(self):
        # Close connections to db when destroy this object
        if self._mapper is not None:
            self.close()
        
    def close(self):
        if self._mapper is not None:
            self._mapper.close()
            self._mapper = None
        
    def clear(self):
        self._mapper.clear()
        self._idCount = 0
        self._size.set(0)
         
    def append(self, item):
        """ Add an item to the set.
        If the item has already an id, use it.
        If not, keep a counter with the max id
        and assign the next one.
        """
        # The _idCount and _size properties work fine
        # under the assumption that once a Set is stored,
        # then it is read-only (no more appends).
        #
        # Anyway, this can be easily changed by updating
        # both from the underlying sqlite when reading a set.

        if not item.hasObjId():
            self._idCount += 1
            item.setObjId(self._idCount)
        else:
            self._idCount = max(self._idCount, item.getObjId())
        self._insertItem(item)
        self._size.increment()

    def _insertItem(self, item):
        self._getMapper().insert(item)
        
    def update(self, item):
        """ Update an existing item. """
        self._getMapper().update(item)
                
    def __str__(self):
        return "%-20s (%d items%s)" % (self.getClassName(), self.getSize(),
                                       self._appendStreamState())

    def _appendStreamState(self):
        return "" if self.isStreamClosed() else ", open set"

    def getSubset(self, n):
        """ Return a subset of n element, making a clone of each. """
        subset = []
        for i, item in enumerate(self):
            subset.append(item.clone())
            if i == n:
                break
        return subset
    
    def setRepresentative(self, representative):
        self._representative = representative
    
    def getRepresentative(self):       
        return self._representative
    
    def hasRepresentative(self):
        """ Return true if have a representative image. """
        return self._representative is not None

    def equalItemAttributes(self, other, ignore=[], verbose=False):
        """Compare that all items in self and other
        return True for equalAttributes.
        """
        return all(x.getObjId() == y.getObjId() and
                   x.equalAttributes(y, ignore=ignore, verbose=verbose)
                   for x, y in zip(self, other))
        
    def hasProperty(self, key):
        return self._getMapper().hasProperty(key)
        
    def getProperty(self, key, defaultValue=None):
        return self._getMapper().getProperty(key, defaultValue)
    
    def loadProperty(self, propertyName, defaultValue=None):
        """ Get the value of a property and
        set its value as an object attribute.
        """
        self.setAttributeValue(propertyName, self.getProperty(propertyName, defaultValue))
        
    def loadAllProperties(self):
        """ Retrieve all properties stored by the mapper. """
        for key in self._getMapper().getPropertyKeys():
            if key != 'self':
                self.loadProperty(key)
        
    def getIdSet(self):
        """ Return a Python set object containing all ids. """
        s = set()
        for item in self.iterItems():
            s.add(item.getObjId())
        return s
    
    def getFiles(self):
        files = set()
        if self.getFileName():
            files.add(self.getFileName())
        return files
    
    def getStreamState(self):
        return self._streamState.get()
    
    def setStreamState(self, newState):
        self._streamState.set(newState)
    
    def isStreamOpen(self):
        return self.getStreamState() == self.STREAM_OPEN
    
    def isStreamClosed(self):
        return self.getStreamState() == self.STREAM_CLOSED 
    
    def enableAppend(self):
        """ By default, when a Set is loaded, it is opened
        in read-only mode, so no new insertions are allowed.
        This function will allow to append more items
        to an existing set.
        """
        self._getMapper().enableAppend()

    @classmethod
    def create(cls, path, template=None, suffix='', **kwargs):
        """ Create a set and set the filename using the suffix.
        If the file exists, it will be delete. """

        if template is None:
            template = cls.FILE_TEMPLATE_NAME

        setFn = os.path.join(path, template % suffix)
        # Close the connection to the database if
        # it is open before deleting the file
        utils.cleanPath(setFn)
        import pyworkflow.mapper as pwmapper
        pwmapper.SqliteDb.closeConnection(setFn)
        setObj = cls(filename=setFn, **kwargs)
        return setObj

def ObjectWrap(value):
    """This function will act as a simple Factory
    to create objects from Python basic types"""
    t = type(value)
    if issubclass(t, Object):
        return value
    if t is int:
        return Integer(value)
    if t is bool:
        return Boolean(value)
    if t is float:
        return Float(value)
    if t is list:
        o = CsvList()
        o.set(value)
        return o
    if t is None:
        return None
    # If it is str, unicode or unknown type, convert to string
    return String(value)
         

class Dict(dict):
    """ Simple wrapper around dict class to have a default value. """
    def __init__(self, default=None):
        self._default = default
        dict.__init__(self)

    def __getitem__(self, key):
        """ Get the image with the given id. """
        return self.get(key, self._default)

    def __contains__(self, item):
        return True


# Define a global dict with all basic defined objects
OBJECTS_DICT = getSubclasses(Object, globals())



import os
from urllib.request import Request
import webbrowser
import json

from . import config


class WorkflowRepository(object):
    """ Manager to communicate with the workflow repository services.
    It will provide functions to:
    - Search workflows (open the url in a browser).
    - Upload a given workflow json file.
    """
    def __init__(self,
                 repositoryUrl=config.WORKFLOW_REPOSITORY_SERVER,
                 uploadFileSuffix=config.WORKFLOW_PROG_STEP1,
                 uploadMdSuffix=config.WORKFLOW_PROG_STEP2):
        self._url = repositoryUrl
        self._uploadFileUrl = repositoryUrl + uploadFileSuffix
        self._uploadMdUrl = repositoryUrl + uploadMdSuffix

    def search(self):
        """ Open the repository URL in a webbrowser. """
        webbrowser.open(self._url)

    def upload(self, jsonFileName):
        """ Upload a given workflow providing the path ot the json file.

        First the file is uploaded, then the metadata is uploaded.
        The script uploads the file and then opens a browser for the metadata
        Note that the two steps are needed since noinitial value can be passed
        to a file field. poster module is needed. Poster is pure python
        so it may be added to the directory rather than installed if needed.

        The server is django a uses filefield and csrf_exempt.
        csrf_exempt disable csrf checking. filefield
        """
        # json file to upload
        params = dict(json=open(jsonFileName, 'rb'))

        # we are going to upload a file so this is a multipart
        # connection
        import poster.streaminghttp
        datagen, headers = poster.encode.multipart_encode(params)
        opener = poster.streaminghttp.register_openers()

        # create request and connect to server
        response = opener.open(Request(self._uploadFileUrl, datagen, headers))
        # server returns dictionary as json with remote name of the saved file
        _dict = json.loads(response.read())

        version = os.environ['SCIPION_VERSION']
        # version hack end

        fnUrl = "?jsonFileName=%s&versionInit=%s" % (_dict['jsonFileName'],
                                                     version)  # use GET
        # open browser to fill metadata, fileNAme will be saved as session
        # variable. Note that I cannot save the file nave in the
        # session in the first connection because the browser changes
        # from urlib2 to an actual browser
        # so sessions are different
        webbrowser.open(self._uploadMdUrl + fnUrl)

    # try:
    #     jsonFileName = os.path.join(tempfile.mkdtemp(), 'workflow.json')
    #     project.exportProtocols(protocols, jsonFileName)
    #     uploadWorkflow(upload_file_url, upload_metadata_url, jsonFileName)
    #     return 0
    # except Exception as ex:
    #     #self.windows.showError(str(ex))
    #     return(ex)



#!/usr/bin/env python
# coding: latin-1
"""
Created on Mar 25, 2014

@author: airen
@author: roberto.marabini
"""



from subprocess import Popen
from io import StringIO

from pyworkflow import APPS
from pyworkflow.utils.process import killWithChilds
from pyworkflow.tests import *
from pyworkflow.utils import utils, prettyDict
from pyworkflow.utils import ProgressBar


class TestBibtex(BaseTest):
    """ Some minor tests to the bibtexparser library. """

    @classmethod
    def setUpClass(cls):
        setupTestOutput(cls)
        
    def test_Parsing(self):
        bibtex = """

@article{delaRosaTrevin2013,
title = "Xmipp 3.0: An improved software suite for image processing in electron microscopy ",
journal = "Journal of Structural Biology ",
volume = "184",
number = "2",
pages = "321 - 328",
year = "2013",
issn = "1047-8477",
doi = "http://dx.doi.org/10.1016/j.jsb.2013.09.015",
url = "http://www.sciencedirect.com/science/article/pii/S1047847713002566",
author = "J.M. de la Rosa-Trevín and J. Otón and R. Marabini and A. Zaldívar and J. Vargas and J.M. Carazo and C.O.S. Sorzano",
keywords = "Electron microscopy, Single particles analysis, Image processing, Software package "
}

@incollection{Sorzano2013,
title = "Semiautomatic, High-Throughput, High-Resolution Protocol for Three-Dimensional Reconstruction of Single Particles in Electron Microscopy",
booktitle = "Nanoimaging",
year = "2013",
isbn = "978-1-62703-136-3",
volume = "950",
series = "Methods in Molecular Biology",
editor = "Sousa, Alioscka A. and Kruhlak, Michael J.",
doi = "10.1007/978-1-62703-137-0_11",
url = "http://dx.doi.org/10.1007/978-1-62703-137-0_11",
publisher = "Humana Press",
keywords = "Single particle analysis; Electron microscopy; Image processing; 3D reconstruction; Workflows",
author = "Sorzano, CarlosOscar and Rosa Trevín, J.M. and Otón, J. and Vega, J.J. and Cuenca, J. and Zaldívar-Peraza, A. and Gómez-Blanco, J. and Vargas, J. and Quintana, A. and Marabini, Roberto and Carazo, JoséMaría",
pages = "171-193",
}
"""

        prettyDict(utils.parseBibTex(bibtex))
        

def wait(condition, timeout=30):
    """ Wait until "condition" returns False or return after timeout (seconds)
    param"""
    t0 = time.time()

    while condition():
        time.sleep(1)

        # Check timeout
        tDelta = time.time()

        if tDelta - t0 >= timeout:
            print("Wait timed out after ", timeout, " seconds")
            return


class TestProccess(BaseTest):
    """ Some tests for utils.process module. """

    @classmethod
    def setUpClass(cls):
        setupTestOutput(cls)
        
    def test_Process(self):
        prog = pw.join(APPS, 'pw_sleep.py')
        p = Popen('python %s 500' % prog, shell=True)
        print("pid: %s" % p.pid)
        time.sleep(5)
        killWithChilds(p.pid)


class TestGetListFromRangeString(BaseTest):

    def test_getListFromRangeString(self):
        inputStrings = ["1,5-8,10",         "2,6,9-11",        "2 5, 6-8"]
        outputLists = [[1, 5, 6, 7, 8, 10], [2, 6, 9, 10, 11], [2, 5, 6, 7, 8]]

        for s, o in zip(inputStrings, outputLists):
            self.assertEqual(o, pwutils.getListFromRangeString(s))
            # Check that also works properly with spaces as delimiters
            s2 = s.replace(',', ' ')
            self.assertEqual(o, pwutils.getListFromRangeString(s2))


class TestProgressBar(unittest.TestCase):

    def caller(self, total, step, fmt, resultGold):
        ti = time.time()
        result = StringIO()
        pb = ProgressBar(total=total, fmt=fmt, output=result,
                         extraArgs={'objectId': 33})

        pb.start()
        for i in range(total):
            if i % step == 0:
                pb.update(i+1)
        pb.finish()
        self.assertEqual(resultGold.strip(), result.getvalue().strip())
        result.close()
        tf = time.time()
        print("%d iterations in %f sec" % (total, tf - ti))

    def test_dot(self):
        total = 1000000
        step = 10000
        ratio = int(total/step)
        resultGold = '.' * (ratio+1)
        self.caller(total=total, step=step,
                    fmt=ProgressBar.DOT, resultGold=resultGold)

    def test_default(self):
        total = 3
        step = 1
        resultGold = ('\rProgress: [                                        ] '
                      '  0%\rProgress: [=============                         '
                      '  ]  33%\rProgress: [==========================        '
                      '      ]  66%\rProgress: [=============================='
                      '==========] 100%')
        self.caller(total=total, step=step,
                    fmt=ProgressBar.DEFAULT, resultGold=resultGold)

    def test_full(self):
        total = 3
        step = 1
        resultGold = ('\r[                                        ] 0/3 (  0%)'
                      ' 3 to go\r[=============                           ] '
                      '1/3 ( 33%) 2 to go\r[==========================      '
                      '        ] 2/3 ( 66%) 1 to go\r[======================'
                      '==================] 3/3 (100%) 0 to go')
        self.caller(total=total, step=step,
                    fmt=ProgressBar.FULL, resultGold=resultGold)

    def test_objectid(self):
        total = 3
        step = 1
        ratio = int(total/step)
        resultGold = ('\r[                                        ] 0/3 (  0%)'
                      ' (objectId=33)\r[=============                         '
                      '  ] 1/3 ( 33%) (objectId=33)\r[========================'
                      '==              ] 2/3 ( 66%) (objectId=33)\r[=========='
                      '==============================] 3/3 (100%) '
                      '(objectId=33)')
        self.caller(total=total, step=step,
                    fmt=ProgressBar.OBJID, resultGold=resultGold)


if __name__ == '__main__':
    unittest.main()        



"""
Created on Apr 8, 2013

@author: antonio
"""

# This file is not really being used, and its tests cannot run because
# of this. So far we are not really using paramiko, and so all its
# dependencies are broken.
#
# I am commenting out all of them today. If used in the future, please
# fix. If not, please remove this file (remote.py) and test_remote.py
#
# Jordi. June 2014.


import os
from os.path import join
import shutil
import socket
# import paramiko
import hashlib

from pyworkflow.utils import exists
from pyworkflow.utils.path import missingPaths, makeFilePath
from pyworkflow.utils.log import ScipionLogger

LOCAL_USER_AND_HOST = ''
SSH_PORT = 22
PAIRS_SEPARATOR = ':'

log = ScipionLogger()


class FileTransfer:
    
    ssh = None   
    sftp = None 
    
    def __init__(self):
        raise NotImplementedError('No dependence on paramiko yet')
        #
        # # Default ssh session options.
        # self.ssh = paramiko.SSHClient()
        # self.ssh.load_system_host_keys()
        # self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    def transferFiles(self,
                      filePaths, 
                      hostsPasswords, 
                      gatewayHosts=None,
                      numberTrials=1,
                      forceOperation=False,
                      operationId=1):
        """
        filePaths -- Files dictionary with this format: "userName@hostName:absolute_file_path": ["userName1@hostName1:absolute_file_path1", "userName2@hostName2:absolute_file_path2"]
        Key is the source file path and value the target file paths.
        gatewayHosts -- Gateway hosts dictionary with this format: "{userName1@hostName1:userName2@hostName2":["userName3@hostName3","userName4@hostName4"]}
        hostsPasswords -- Passwords needed to connect to involved hosts with this format: "userName@hostName":"hostPassword"
        numberTrials -- Number of trials in error cases.
        forceOperation -- Flag to indicate if, when an error happens and number of trials is exceeded, the operation must continue with the rest of files.        
        operationId -- Operation identifier.
        """
        classifiedFiles = self.__classifyFilePaths(filePaths)
        
        for userAndHostPairs, cursorFilePaths in classifiedFiles.items():
            gatewayHostsCredentials = None
            if gatewayHosts is not None and userAndHostPairs in gatewayHosts:
                gatewayHostsCredentials = gatewayHosts[userAndHostPairs]
            self.__sendFilesBetweenPairs(userAndHostPairs, 
                                         cursorFilePaths,
                                         hostsPasswords, 
                                         gatewayHostsCredentials, 
                                         numberTrials,                        
                                         forceOperation,
                                         operationId) 
    
    def copyFiles(self,
                  filePaths,                        
                  numberTrials=1,
                  forceOperation=False,
                  operationId=1):
        """
        filePaths -- Files dictionary with this format: "source_file_path": "target_file_path"
        forceOperation -- Flag to indicate if, when an error happens and number of trials is exceeded, the operation must continue with the rest of files.        
        operationId -- Operation identifier.
        """
        for sourceFilePath, targetFilePath in filePaths.items():
            self.__copyLocalFile(sourceFilePath, targetFilePath)
    
    def transferFilesTo(self,
                        filePaths,
                        hostName,
                        userName,
                        hostPassword,
                        gatewayHosts=None,
                        numberTrials=1,
                        forceOperation=False,
                        operationId=1):
        """
        filePaths -- Files dictionary with this format: "source_file_path": "target_file_path"
        hostName -- Remote host to transfer files.
        username -- User name for remote host.
        hostsPassword -- Passwords needed to connect to involved host.
        gatewayHosts -- Gateway hosts List with this format: ["userName1@hostName1","userName2@hostName2"]
        numberTrials -- Number of trials in error cases.
        forceOperation -- Flag to indicate if, when an error happens and number of trials is exceeded, the operation must continue with the rest of files.        
        operationId -- Operation identifier.
        """
        # Create ssh session to remote host
        log.info("Connecting to: " + userName + "@" + hostName)
        self.ssh.connect(hostName, SSH_PORT, userName, hostPassword)
        self.sftp = self.ssh.open_sftp()
        for sourceFilePath, targetFilePath in filePaths.items():
            self.__sendLocalFile(sourceFilePath, targetFilePath, gatewayHosts, self.sftp)
        self.ssh.close()
        self.sftp.close()  
            
    def transferFilesFrom(self,
                          filePaths,
                          hostName,
                          userName,
                          hostPassword,
                          gatewayHosts=None,
                          numberTrials=1,
                          forceOperation=False,
                          operationId=1):
        """
        filePaths -- Files dictionary with this format: "source_file_path": "target_file_path"
        hostName -- Remote host to transfer files.
        username -- User name for remote host.
        hostsPassword -- Passwords needed to connect to involved host.
        gatewayHosts -- Gateway hosts List with this format: ["userName1@hostName1","userName2@hostName2"]
        numberTrials -- Number of trials in error cases.
        forceOperation -- Flag to indicate if, when an error happens and number of trials is exceeded, the operation must continue with the rest of files.        
        operationId -- Operation identifier.
        """
        # Create ssh session to remote host
        log.info("Connecting to: " + userName + "@" + hostName)
        self.ssh.connect(hostName, SSH_PORT, userName, hostPassword)
        self.sftp = self.ssh.open_sftp()
        for sourceFilePath, targetFilePath in filePaths.items():
            self.__getRemoteFile(sourceFilePath, targetFilePath, gatewayHosts, self.sftp)
        self.ssh.close()
        self.sftp.close()  
                
    def deleteFiles(self,
                    filePaths,                    
                    hostsPasswords, 
                    gatewayHosts=None, 
                    numberTrials=1, 
                    forceOperation=False, 
                    operationId=1):
        """
        Delete a list of file paths.
        filepaths -- List of file paths to remove with this format: "userName@hostName:absolute_file_path"
        gatewayHosts -- Gateway hosts dictionary with this format: "userName1@hostName1:userName2@hostName2":"userName@hostName"
        hostsPasswords -- Passwords needed to connect to involved hosts with this format: "userName@hostName":"hostPassword"
        numberTrials -- Number of trials in error cases.
        forceOperation -- Flag to indicate if, when an error happens and number of trials is exceeded, the operation must continue with the rest of files.
        operationId -- Operation identifier.
        """
        
        # As we are going to create a session for each target host we must get different target hosts.
        userAndHosts = self.__getDiferentHostsFromFilePaths(filePaths)
        for userAndHost in userAndHosts:
            if not self.__isLocalCredential(userAndHost):
                resultFilePaths = self.__getFilePaths(filePaths, userAndHost)
                # Recover host credentials to remove files
                userName = self.__getUserAndHost(userAndHost)[0]
                hostName = self.__getUserAndHost(userAndHost)[1]
                hostPassword = hostsPasswords[userAndHost]
                # Create ssh session to remote host
                log.info("Connecting to: " + userName + "@" + hostName)
                self.ssh.connect(hostName, SSH_PORT, userName, hostPassword)
                self.sftp = self.ssh.open_sftp()
                for resultFilePath in resultFilePaths:
                    filePath = self.__getLocationAndFilePath(resultFilePath)[1]
                    log.info("Deleting file " + filePath)
                    self.sftp.remove(filePath)
                self.ssh.close()
                self.sftp.close()  
            else:
                pass
            
    def deleteDirectories(self, 
                          directoryPaths,                    
                          hostsPasswords, 
                          gatewayHosts=None, 
                          numberTrials=1, 
                          forceOperation=False, 
                          operationId=1):
        """
        Delete a list of directory paths.
        directoryPaths -- List of file paths to remove with this format: "userName@hostName:absolute_directory_path"
        gatewayHosts -- Gateway hosts dictionary with this format: "userName1@hostName1:userName2@hostName2":"userName@hostName"
        hostsPasswords -- Passwords needed to connect to involved hosts with this format: "userName@hostName":"hostPassword"
        numberTrials -- Number of trials in error cases.
        forceOperation -- Flag to indicate if, when an error happens and number of trials is exceeded, the operation must continue with the rest of files.
        operationId -- Operation identifier.
        """
        
        # As we are going to create a session for each target host we must get different target hosts.
        userAndHosts = self.__getDiferentHostsFromFilePaths(directoryPaths)
        for userAndHost in userAndHosts:
            if not self.__isLocalCredential(userAndHost):
                resultDirectoryPaths = self.__getFilePaths(directoryPaths, userAndHost)
                # Recover host credentials to remove files
                userName = self.__getUserAndHost(userAndHost)[0]
                hostName = self.__getUserAndHost(userAndHost)[1]
                hostPassword = hostsPasswords[userAndHost]
                # Create ssh session to remote host
                log.info("Connecting to: " + userName + "@" + hostName)
                self.ssh.connect(hostName, SSH_PORT, userName, hostPassword)
                self.sftp = self.ssh.open_sftp()
                for resultDirectoryPath in resultDirectoryPaths:
                    directoryName = self.__getLocationAndFilePath(resultDirectoryPath)[1]
                    log.info("Deleting directory " + directoryName)
                    self.sftp.rmdir()
                self.ssh.close()
                self.sftp.close()  
            else:
                pass
        
    def checkFiles(self, 
                   filePaths,                   
                   hostsPasswords,
                   gatewayHosts=None, 
                   numberTrials=1, 
                   forceOperation=False,
                   operationId=1):
        """
        Check if file paths exists.
        filepaths -- List of file paths to check with this format: "userName@hostName:absolute_file_path"
        gatewayHosts -- Gateway hosts dictionary with this format: "userName1@hostName1:userName2@hostName2":"userName@hostName"
        hostsPasswords -- Passwords needed to connect to involved hosts with this format: "userName@hostName":"hostPassword"
        numberTrials -- Number of trials in error cases.
        forceOperation -- Flag to indicate if, when an error happens and number of trials is exceeded, the operation must continue with the rest of files.
        operationId -- Operation identifier.
        returns -- List of not located file paths.
        """
        returnFilePaths = []
        # As we are going to create a session for each target host we must get different target hosts.
        userAndHosts = self.__getDiferentHostsFromFilePaths(filePaths)
        for userAndHost in userAndHosts:
            resultFilePaths = self.__getFilePaths(filePaths, userAndHost)
            if self.__isLocalCredential(userAndHost):
                for resultFilePath in resultFilePaths:
                    filePath = self.__getLocationAndFilePath(resultFilePath)[1]
                    log.info("Checking: " + filePath)
                    if len(missingPaths(filePath)) != 0:
                        returnFilePaths.append(filePath)
                        log.info("Check fail!!")
            else:
                
                # Recover host credentials
                userName = self.__getUserAndHost(userAndHost)[0]
                hostName = self.__getUserAndHost(userAndHost)[1]
                hostPassword = hostsPasswords[userAndHost]
                # Create ssh session to remote host
                log.info("Connecting to: " + userName + "@" + hostName)
                self.ssh.connect(hostName, SSH_PORT, userName, hostPassword)
                self.sftp = self.ssh.open_sftp()
                for resultFilePath in resultFilePaths:
                    filePath = self.__getLocationAndFilePath(resultFilePath)[1]
                    log.info("Checking: " + filePath)
                    try:
                        self.sftp.lstat(filePath)                    
                    except IOError:
                        returnFilePaths.append(resultFilePath)
                        log.info("Check fail!!")
                self.ssh.close()
                self.sftp.close()  
        return returnFilePaths
    
    def checkOneHostFiles(self, 
                          filePaths,
                          hostName,
                          userName,
                          hostsPassword,
                          gatewayHosts=None,
                          numberTrials=1,
                          forceOperation=False,
                          operationId=1):
        """
        Check if file paths exists.
        filepaths -- List of file paths to check with this format: ["absolute_file_path1","absolute_file_path2"]
        gatewayHosts -- Gateway hosts dictionary with this format: "userName1@hostName1:userName2@hostName2":"userName@hostName"
        hostsPasswords -- Passwords needed to connect to involved hosts with this format: "userName@hostName":"hostPassword"
        numberTrials -- Number of trials in error cases.
        forceOperation -- Flag to indicate if, when an error happens and number of trials is exceeded, the operation must continue with the rest of files.
        operationId -- Operation identifier.
        returns -- List of not located file paths.
        """
        returnFilePaths = []
        print("**************************************** CHECKING******************************************")
        log.info("Connecting to: " + userName + "@" + hostName)
        self.ssh.connect(hostName, SSH_PORT, userName, hostsPassword)
        self.sftp = self.ssh.open_sftp()
        
        isLocalHost = self.__isLocalHost(hostName)
        
        for fileName in filePaths:
            log.info("Checking: " + fileName)
            if isLocalHost:
                if len(missingPaths(fileName)) != 0:
                    returnFilePaths.append(fileName)
                    log.info("Check fail!!")
            else:
                try:
                    self.sftp.lstat(fileName)                    
                except IOError:
                    returnFilePaths.append(fileName)
                    log.info("Check fail!!")
        self.ssh.close()
        self.sftp.close()
        return returnFilePaths
    
    def __classifyFilePaths(self, filePaths):
        """
        Classify file paths to send depending on the source and target user/host involved credentials.
        filePaths -- Files dictionary with this format: "userName@hostName:absolute_file_path": [userName1@hostName1:absolute_file_path1, userName2@hostName2:absolute_file_path2]
        Key is the source file path and value the target file paths.
        returns -- Dictionary with file paths ordered by the two user/host credentials that are involved. The dictionary will have this structure:
        {"userName1@hostName1:userName2@hostName2" : {"filePath1":"filePath2"}
        """
        result = {}
        
        for sourcePath, targetPaths in filePaths.items():
            for targetPath in targetPaths:
                sourceParts = self.__getLocationAndFilePath(sourcePath)
                targetParts = self.__getLocationAndFilePath(targetPath) 
                sourceUserAndHost = sourceParts[0]
                targetUserAndHost = targetParts[0]
                sourceFilePath = sourceParts[1]
                targetFilePath = targetParts[1]
                resultKey = sourceUserAndHost + PAIRS_SEPARATOR + targetUserAndHost
                auxDict = {}
                if resultKey in result:
                    auxDict = result[resultKey]
                    if sourceFilePath in auxDict:
                        # This will not happen in Scipion because one source path is not going
                        # to be sent to different file paths in the same machine.
                        msg = 'File ' + sourcePath + ' can not be sent to ' + auxDict[sourcePath] + " and to " + targetPath + ' because they are in the same machine.'
                        raise Exception(msg)
                auxDict[sourceFilePath] = targetFilePath
                result[resultKey] = auxDict
        
        return result
            
    def __sendFilesBetweenPairs(self, 
                                userAndHostPairs, 
                                filePaths,
                                hostsPasswords, 
                                gatewayHosts=None,
                                numberTrials=1,
                                forceOperation=False,
                                operationId=1):
        """
        filePaths -- Dictionary with this structure: {"filePath1":"filePath2"}
        gatewayHosts -- Gateway hosts List with this format: ["userName1@hostName1","userName2@hostName2"] 
        TODO: Get the target directories to check it existence and create them only once.
        """     
        pairParts = self.__getUserAndHosts(userAndHostPairs)
        sourceCredentials = pairParts[0]
        targetCredentials = pairParts[1]
        # We see what type of sending operation is.        
        if self.__isLocalCredential(sourceCredentials):
            if self.__isLocalCredential(targetCredentials):
                self.copyFiles(filePaths, numberTrials, forceOperation, operationId)
            else:
                targetUserAndHost = self.__getUserAndHost(targetCredentials)
                self.transferFilesTo(filePaths, targetUserAndHost[1],
                                     targetUserAndHost[0], hostsPasswords[targetCredentials],
                                     gatewayHosts, numberTrials, forceOperation, operationId)
        else:
            sourceUserAndHost = self.__getUserAndHost(sourceCredentials)
            if self.__isLocalCredential(targetCredentials):
                self.transferFilesFrom(filePaths, sourceUserAndHost[1], sourceUserAndHost[0],
                                       hostsPasswords[sourceCredentials], gatewayHosts,
                                       numberTrials, forceOperation, operationId)
            else:
                pass            
            
    def __getUserAndHosts(self, userAndHostPairs):
        """
        Separate user and host pair in their individuals.
        userAndHostPairs -- User and host pair: "userName1@hostName1:userName2@hostName2"
        returns -- Spplited pairs: ["userName1@hostName1", "userName2@hostName2"]
        """
        return userAndHostPairs.split(PAIRS_SEPARATOR)
    
    def __getUserAndHost(self, userAndHost):
        """
        Function to get the user and the host name from 'userName@hostName' string
        Returns: Tuple with ('userName', 'hostName')
        """
        return userAndHost.split("@")
    
    def __getLocationAndFilePath(self, locationAndFile):
        """
        Function to get the user and the userName@hostName and file path from 'userName@hostname:filePath' string
        Returns: Tuple with ('userName@hostName', 'filePath')
        """
        if ":" in locationAndFile:
            auxLocationAndFile = locationAndFile.split(":", 1)
            if self.__isLocalCredential(auxLocationAndFile[0]):
                auxLocationAndFile[0] = LOCAL_USER_AND_HOST  # Ease classification and other operations
        else:
            auxLocationAndFile = list()
            auxLocationAndFile.append(LOCAL_USER_AND_HOST)
            auxLocationAndFile.append(locationAndFile)
        
        return auxLocationAndFile
    
    def __isLocalHost(self, hostName):
        """
        Checks if one host name is local machine name.
        No name or empty name means local host.
        """
        if (hostName is None or
                hostName == LOCAL_USER_AND_HOST):
            return True
        elif socket.gethostname() == hostName:
            return True
        else:
            return False

    def __isLocalCredential(self, userAndHost):
        """
        Checks if one userName@hostName credential is about local machine.
        """
        if (userAndHost is None or
            userAndHost == LOCAL_USER_AND_HOST):
            return True
        else:
            hostName = self.__getUserAndHost(userAndHost)[1]
            return self.__isLocalHost(hostName)
        
    def __getDiferentHostsFromFilePaths(self, filePaths):
        """
        Gets different userName@hostName credentials of files list.
        files -- File list with this format: userName1@hostName1:absolute_file_path1
        returns -- List with different files userName@hostName credentials.
        """
        differentHosts = []
        for filePath in filePaths:
            userAndHost = self.__getLocationAndFilePath(filePath)[0]
            if self.__isLocalCredential(userAndHost):
                userAndHost = LOCAL_USER_AND_HOST  # To simplify code for local credentials.
            if userAndHost not in differentHosts:
                differentHosts.append(userAndHost)
        return differentHosts
    
    def __getFilePaths(self, filePaths, userAndHost):
        """
        Gets a file path from a list of target file paths for the given user and host.
        filePaths -- List of files with this format: "useName@hostName:/filePath"
        userAndHost -- String with this format. "userName@hostName" 
        returns -- Target file path list ("useName@hostName:/filePath") for the given user and host.
        """  
        resultFilePaths = []             
        for filePath in filePaths:
            locationAndFilePath = self.__getLocationAndFilePath(filePath)
            if locationAndFilePath[0] == userAndHost:
                resultFilePaths.append(filePath)
        return resultFilePaths
    
    def __copyLocalFile(self, sourceFilePath, targetFilePath):
        """
        Send local file to remote machine
        sourceFilePath -- Source file path (/file path/...).
        targetFilePath -- Target file path (/file path/...).
        """
        log.info("Copying " + sourceFilePath + " to " + targetFilePath)
        # Check if file already existsFilePath and it is up to date
        existsFilePath = False
        if exists(targetFilePath):
            if self.__getLocalSHA1(sourceFilePath) == self.__getLocalSHA1(targetFilePath):
                existsFilePath = True
                log.info(targetFilePath + " already existed")
        if not existsFilePath:
            makeFilePath(targetFilePath)
            shutil.copy2(sourceFilePath, targetFilePath)
        
    def __sendLocalFile(self, sourceFilePath, targetFilePath, gatewayHosts, sftp):
        """
        Send local file to remote machine
        sourceFilePath -- Source file path (/file path/...).
        targetFilePath -- Target file path (/file path/...).
        sftp -- sftp connection.
        """
        log.info("Sending " + sourceFilePath + " to " + targetFilePath)
        # Check if file already existsFilePath and it is up to date
        existsFilePath = False
        if self.__existsRemotePath(targetFilePath, sftp):
            if self.__getLocalSHA1(sourceFilePath) == self.__getRemoteSHA1(targetFilePath, self.ssh):
                existsFilePath = True
                log.info(targetFilePath + " already existed")
        if not existsFilePath:
            self.__createRemoteFolderForFile(targetFilePath, sftp)
            try:
                sftp.put(sourceFilePath, targetFilePath)
            except IOError as err:
                log.error("Fail sending local file " + sourceFilePath +
                          " to remote file " + targetFilePath + " - " + str(err))
                raise
        
    def __getRemoteFile(self, sourceFilePath, targetFilePath, gatewayHosts, sftp):
        """
        Send local file to remote machine
        sourceFilePath -- Source file path (/file path/...).
        targetFilePath -- Target file path (/file path/...).
        sftp -- sftp connection.
        """
        log.info("Getting " + sourceFilePath + " to " + targetFilePath)
        # Check if file already existsFilePath and it is up to date
        existsFilePath = False
        if exists(targetFilePath):
            if self.__getRemoteSHA1(sourceFilePath, self.ssh) == self.__getLocalSHA1(targetFilePath):
                existsFilePath = True
                log.info(targetFilePath + " already existed")
        if not existsFilePath:
            makeFilePath(targetFilePath)
            try:
                sftp.get(sourceFilePath, targetFilePath)
            except IOError as err:
                log.error("Fail getting remote file " + sourceFilePath +
                          " to local file " + targetFilePath + " - " + str(err))
                raise
    
    def __createRemoteFolderForFile(self, filePath, sftp):
        """
        Create folder for file in remote host defined by sfpt.
        filePath -- File path which parent directory we must create (/file path/...).
        sftp -- Remote sftp session.
        """
        filePathParentDirectory = os.path.dirname(filePath)
        self.__mkdirP(filePathParentDirectory, sftp)

    def __mkdirP(self, remoteDirectory, sftp):
        """
        Create remote folder structure creating all non-existent folders.
        remoteDirectory -- Remote directory to create.
        sftp -- Remote sftp session.
        """
        if not self.__existsRemotePath(remoteDirectory, sftp):
            self.__mkdirP(os.path.dirname(remoteDirectory), sftp)
            sftp.mkdir(remoteDirectory) 
            
    def __getLocalSHA1(self, filePath):
        return hashlib.sha1(open(filePath, 'rt').read()).hexdigest()

    def __getRemoteSHA1(self, filePath, ssh):
        stdin, stdout, stderr = ssh.exec_command("sha1sum '" + filePath + "'")
        return stdout.readlines()[0].split()[0] 
            
    def __existsRemotePath(self, path, sftp):
        try:
            sftp.lstat(path)
            return True            
        except IOError:
            return False


# -----------------------------------------------------------------------------
#                AUXILIARY FUNCTIONS
# -----------------------------------------------------------------------------

def isRemoteDir(sftp, path):
    """ Check if one remote directory exists
    Params:
        sftp: Sftp session.
        path: Path to check.
    Returns: True if the given path is a directory, false if it is not a directory.
    """
    raise NotImplementedError('No dependence on paramiko yet')
    #
    # try:
    #     sftp.chdir(path)
    #     return True
    # except (IOError, paramiko.SFTPError):
    #     return False


def getRemoteFolderFiles(hostName, userName, password, folderPath, recursive=True):
    """ Recover all files in the given folder.
    Params:
        hostName: Remote host name.
        userName: User name.
        password: Password.
        folderPath: Folder to get files.
        recursive: if True go recursively inside other subfolders.
    Returns: List of files.
    """    
    raise NotImplementedError('No dependence on paramiko yet')
    #
    # # Default ssh session options.
    #
    # ssh = paramiko.SSHClient()
    # ssh.load_system_host_keys()
    # ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    # ssh.connect(hostName, 22, userName, password)
    # sftp = ssh.open_sftp()
    # remoteFiles = getRemoteFiles(sftp, folderPath, recursive)
    # sftp.close()
    # ssh.close()
    # return remoteFiles


def getRemoteFiles(sftp, folderPath, recursive=True):
    """ Recover all files in the given folder.
    Params:
        sftp: Sftp session.
        folderPath: Folder to get files.
        recursive: if True go recursively inside other subfolders.
    Returns: List of files.
    """    
    filePathList = sftp.listdir(folderPath)
    resultFilePathList = []
    for filePath in filePathList:
        if isRemoteDir(sftp, join(folderPath, filePath)) and recursive:
            resultFilePathList += getRemoteFiles(sftp, join(folderPath, filePath))
        else:
            resultFilePathList.append(join(folderPath, filePath))
    return resultFilePathList


def removeRemoteFolder(hostName, userName, password, folderPath):
    """ Removes a remote folder and all it content.
    Params:
        hostName: Remote host name.
        userName: User name.
        password: Password.
        folderPath: Folder to delete.
    Returns: 
        Tuple with standard input, standard output and error output.
    """
    raise NotImplementedError('No dependence on paramiko yet')
    #
    # ssh = paramiko.SSHClient()
    # ssh.load_system_host_keys()
    # ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    # ssh.connect(hostName, 22, userName, password)
    # return removeTree(ssh, folderPath)


def removeTree(ssh, folderPath):
    """ Removes a remote folder and all it content.
    Params:
        ssh: Ssh session.
        folderPath: Folder to delete.
    Returns: 
        Tuple with standard input, standard output and error output.
    """
    stdin, stdout, stderr = ssh.exec_command("rm -dfr " + folderPath)
    ssh.close()
    return stdin, stdout, stderr


def getFilePathList(filePaths):
    """
    Get target file paths list from target file path dictionary.
    filePaths -- Files dictionary with this format: "userName@hostName:absolute_file_path": ["userName1@hostName1:absolute_file_path1", "userName2@hostName2:absolute_file_path2"]
    returns -- List of target files with this format: "useName@hostName:/filePath".
    """
    resultFilePathList = []
    for filePathList in filePaths.values():
        for filePath in filePathList:
            resultFilePathList.append(filePath)
    return resultFilePathList


if __name__ == '__main__':
    pass



#!/usr/bin/env python
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import os
import sys

from pwem.viewers import DataView
from pyworkflow.gui.browser import FileBrowserWindow

    
def showDir(path):
    window = FileBrowserWindow("Browsing: " + path, path=path) 
    window.show()
    

def showFile(path, viewParams):
    DataView(path, viewParams).show()


if __name__ == '__main__':    
    
    if '-h' in sys.argv or '--help' in sys.argv:
        print("usage: scipion view [file1 file2 file3 ... fileN]")
        
    else:
        if len(sys.argv) == 1:  # no extra arguments, show current directory
            showDir(os.getcwd())
        else:
            args = {'-i': []}
            lastOpt = '-i'
            
            for a in sys.argv[1:]:
                if a.startswith('-'):
                    lastOpt = a
                    if lastOpt not in args:
                        args[lastOpt] = []
                else:
                    args[lastOpt].append(a)
                    
            inputFiles = args['-i']
            del args['-i']
            viewParams = {}
            for k, v in args.items():
                viewParams[k.replace('-', '')] = ' '.join(v)
            
            for fn in inputFiles:
                if os.path.isdir(fn):
                    showDir(fn)
                else:
                    showFile(fn, viewParams)



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This module implement the classes to create plots on xmipp.
"""

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None

from pyworkflow.viewer import View

figureCounter = 0


class Plotter(View):
    """ Create different types of plots using the matplotlib library. """
    backend = None
    
    @classmethod
    def setBackend(cls, value):
        """ Possible values are:
        - TkAgg for Tkinter
        - Agg for non-interactive plots.
        """
        plt.switch_backend(value)
        cls.backend = value
        
    def __init__(self, x=1, y=1, mainTitle="", 
                 figsize=None, dpi=100, windowTitle="",
                 fontsize=8, **kwargs):
        """ This Plotter class has some utilities to create a Matplotlib figure
        and add some plots to it.
        Params:
            x, y: number of rows and colums of the grid for plots.
            mainTitle: figure main title.
            figsize: the size of the figure, if None, it will be guessed from x and y
            dpi: resolution, 100 by default.
            windowTitle: title for the whole windows.
        """
        figure = kwargs.get('figure', None)

        if figure == 'active':
            figure = plt.gcf()
        
        if self.backend is None:
            Plotter.setBackend('Agg')

        plt.style.use(kwargs.get('style', 'default'))

        if figure is None:
            self.tightLayoutOn = True
            
            if figsize is None:  # Set some defaults values
                if x == 1 and y == 1:
                    figsize = (6, 5)
                elif x == 1 and y == 2:
                    figsize = (4, 6)
                elif x == 2 and y == 1:
                    figsize = (6, 4)
                else:
                    figsize = (8, 6)
            
            # Create grid
            # import matplotlib.gridspec as gridspec
            # self.grid = gridspec.GridSpec(x, y)#, height_ratios=[7,4])
            # self.grid.update(left=0.15, right=0.95, hspace=0.25, wspace=0.4)#, top=0.8, bottom=0.2)
            global figureCounter
            figureCounter += 1
            self.figure = plt.figure(figureCounter, figsize=figsize, dpi=dpi)
            # from matplotlib.figure import Figure
            # self.figure = Figure(figsize=figsize, dpi=dpi)
            if mainTitle:
                self.figure.suptitle(mainTitle, fontsize=fontsize + 4)
            if windowTitle:
                self.figure.canvas.set_window_title(windowTitle) 
            self.plot_count = 0
            self.last_subplot = None
            self.plot_yformat = '%1.2e'
        else:
            self.figure = figure
            self.tightLayoutOn = False
            self.plot_count = 0
            
        self.fontsize = fontsize
        self.plot_title_fontsize = fontsize + 4
        self.plot_axis_fontsize = fontsize + 2
        self.plot_text_fontsize = fontsize
        self.gridx = x
        self.gridy = y
            
    def activate(self):
        """ Activate this figure. """
        plt.figure(self.figure.number)
        
    def getCanvas(self):
        return self.figure.canvas
    
    def getFigure(self):
        return self.figure
    
    def showLegend(self, labels, loc='best'):
        leg = self.last_subplot.legend(tuple(labels), loc=loc)
        for t in leg.get_texts():
            t.set_fontsize(self.plot_axis_fontsize)    # the legend text fontsize

    def legend(self, loc='best', **kwargs):
        self.last_subplot.legend(loc=loc, **kwargs)

    def createSubPlot(self, title, xlabel, ylabel, xpos=None, ypos=None,
                      yformat=False, projection='rectilinear'
                      ):
        """
        Create a subplot in the figure.
        You should provide plot title, and x and y axis labels.
        yformat True specified the use of global self.plot_yformat
        Posibles values for projection are:
            'aitoff', 'hammer', 'lambert', 'mollweide', 'polar', 'rectilinear'

        """
        if xpos is None:
            self.plot_count += 1
            pos = self.plot_count
        else:
            pos = xpos + (ypos - 1) * self.gridx
        a = self.figure.add_subplot(self.gridx, self.gridy, pos, projection=projection)
        # a.get_label().set_fontsize(12)
        a.set_title(title, fontsize=self.plot_title_fontsize)
        a.set_xlabel(xlabel, fontsize=self.plot_axis_fontsize)
        a.set_ylabel(ylabel, fontsize=self.plot_axis_fontsize)
            
        if yformat:
            import matplotlib.ticker as ticker
            formatter = ticker.FormatStrFormatter(self.plot_yformat)
            a.yaxis.set_major_formatter(formatter)
        a.xaxis.get_label().set_fontsize(self.plot_axis_fontsize)
        a.yaxis.get_label().set_fontsize(self.plot_axis_fontsize)

        labels = a.xaxis.get_ticklabels() + a.yaxis.get_ticklabels()
        for label in labels:
            label.set_fontsize(self.plot_text_fontsize)  # Set fontsize
            # label.set_text('aa')
        self.last_subplot = a
        self.plot = a.plot
        self.hist = a.hist
        self.scatterP = a.scatter
        self.bar = a.bar
        return a

    def getLastSubPlot(self):
        return self.last_subplot
    
    def createCanvas(self):
        a = self.figure.add_subplot(111, axisbg='g')
        a.set_axis_off()
        self.figure.set_facecolor('white')
        return a

    def getColorBar(self, plot):
        self.tightLayoutOn = False
        cax = self.figure.add_axes([0.9, 0.1, 0.03, 0.8])
        cbar = self.figure.colorbar(plot, cax=cax)
        cbar.ax.invert_yaxis()

    def tightLayout(self):
        if self.tightLayoutOn:
            self.activate()
            plt.tight_layout()
        
    def show(self, interactive=True, block=False):
        self.tightLayout()
        plt.show(block=block)

    def draw(self):
        self.tightLayout()
        self.getCanvas().draw()
        
    def clear(self):
        self.getFigure().clear()
        self.plot_count = 0
        self.last_subplot = None
        
    def savefig(self, *args, **kwargs):
        self.tightLayout()
        self.figure.savefig(*args, **kwargs)
        
    def isClosed(self):
        """ Return true if the figure have been closed. """
        return not plt.fignum_exists(self.figure.number)
    
    def close(self):
        """ Close current Plotter figure. """
        plt.close(self.figure)
        

def getHexColorList(stepColors, colorName='jet'):
    """ Returns a list of hexColor """
    from matplotlib import cm, colors
    
    colorsList = []
    colorMap = cm.get_cmap(colorName)
    ratio = 255.0 / (len(stepColors))
    for index in range(len(stepColors)):
        colorPosition = int(round((index+1) * ratio))
        rgb = colorMap(colorPosition)[:3]
        rgbColor = colors.rgb2hex(rgb)
        colorsList.append(rgbColor)

    return colorsList



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
Several Image tools using Matplotlib.
"""

import tkinter as tk
import matplotlib
import numpy as np

from pyworkflow import TK_GRAY_DEFAULT

try:
    matplotlib.use('TkAgg')
    from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
    import matplotlib.pyplot as plt
except ImportError:
    plt = None

from matplotlib.figure import Figure
import matplotlib.cm as cm
from matplotlib.patches import Wedge


class FigureFrame(tk.Frame):
    """ Create a Tk Frame that will contains a 
    Matplotlib Figure. 
    **kwargs arguments will be passed to Figure constructor.
    Valid options are:
        figsize = (xdim, ydim)
        dpi = value
        frameon = (True|False)
    """
    def __init__(self, parent, **kwargs):
        tk.Frame.__init__(self, parent)
        self.figure = Figure(**kwargs)
        self.canvas = FigureCanvasTkAgg(self.figure, master=self)
        self.canvas.get_tk_widget().grid(column=0, row=0)

    def getFigure(self):
        return self.figure
    
    def getCanvas(self):
        return self.canvas
    
    
class Preview(tk.Frame):
    # def __init__(self, parent, dim, dpi=36, label=None):
    def __init__(self, parent, dim, dpi=36, label=None, col=0, row=0, listenersDict=None):
        tk.Frame.__init__(self, parent)
        self.dim = dim
        self.bg = np.zeros((int(dim), int(dim)), float)
        ddim = dim/dpi
        self.figure = Figure(figsize=(ddim, ddim), dpi=dpi, frameon=False)
        self.canvas = FigureCanvasTkAgg(self.figure, master=self)
        self.canvas.get_tk_widget().grid(column=0, row=0)  # , sticky=(N, W, E, S))
        self.canvas.get_tk_widget().config(bg=TK_GRAY_DEFAULT)
        if label:
            tk.Label(self, text=label).grid(column=0, row=1)
        self._createAxes()

        if listenersDict is not None:
            for bindingKey, callback in listenersDict.items():
                self.canvas.get_tk_widget().bind(bindingKey, callback)

    def setWindowTitle(self, title):
        """ Set window title"""
        self.canvas.set_window_title(title)

    def _createAxes(self):
        """ Should be implemented in subclasses. """
        pass
    
    def _update(self, *args):
        """ Should be implemented in subclasses. """
        pass
    
    def clear(self):
        self._update(self.bg)
        
    def updateData(self, Z):
        self.clear()
        self._update(Z)
    
    
class ImagePreview(Preview):
    def __init__(self, parent, dim, dpi=36, label=None, col=0, listenersDict=None):
        Preview.__init__(self, parent, dim, dpi, label, col, listenersDict=listenersDict)
            
    def _createAxes(self):
        ax = self.figure.add_axes([0, 0, 1, 1], frameon=False)
        self.figureimg = ax.imshow(self.bg, cmap=cm.gray)  # , extent=[-h, h, -h, h])
        ax.set_axis_off()
        self.ax = ax
        
    def _update(self, Z, *args):
        self.figureimg.set_data(Z)
        self.figureimg.autoscale()
        self.figureimg.set(extent=[0, Z.shape[1], 0, Z.shape[0]])
        self.canvas.draw()
        
        
class PsdPreview(Preview):
    def __init__(self, master, dim, lf, hf, dpi=72, label="PSD", listenersDict=None):
        Preview.__init__(self, master, dim, dpi, label, listenersDict=listenersDict)
        self.lf = lf
        self.hf = hf
        if self.ring:
            self.createRing()
        else:
            self.canvas.draw()
                            
    def _createAxes(self):
        ax = self.figure.add_axes([0, 0, 1, 1], frameon=False)
        h = 0.5
        ax.set_xlim(-h, h)
        ax.set_ylim(-h, h)
        ax.tick_params(axis="x", direction="in", pad=-15)
        ax.tick_params(axis="y", direction="in", pad=-22)
        ax.grid(True)
        self.ring = None
        self.img = ax.imshow(self.bg, cmap=cm.gray, extent=[-h, h, -h, h])
        self.ax = ax
        
    def createRing(self, fc=None):
        radius = float(self.hf)
        width = radius - float(self.lf)
        self.ring = Wedge((0, 0), radius, 0, 360, width=width, alpha=0.15, fc=fc)  # Full ring
        self.ax.add_patch(self.ring)
        self.canvas.draw()
        
    def updateFreq(self, lf, hf):
        self.lf = lf
        self.hf = hf
        if self.ring:
            self.ring.remove()
            self.ring = None
        if self.hf:
            self.createRing()
    
    def _update(self, Z):
        if self.ring:
            self.ring.remove()
            self.ring = None
        if self.hf:
            self.createRing()
        self.img.set_data(Z)
        self.img.autoscale()
        self.canvas.draw()
        
        
class MaskPreview(ImagePreview):
    def __init__(self, parent, dim, dpi=36, label=None, col=0, listenersDict=None):
        ImagePreview.__init__(self, parent, dim, dpi, label, col, listenersDict)
        self.ring = None
            
    def updateMask(self, outerRadius, innerRadius=0, fc=None):
        if self.ring is not None:
            self.ring.remove()
        center = self.dim / 2
        width = outerRadius - innerRadius
        self.ring = Wedge((center, center), outerRadius, 0, 360, width=width, alpha=0.15, fc=fc)  # Full ring
        self.ax.add_patch(self.ring)
        self.canvas.draw()
        

def getPngData(filename):
    import matplotlib.image as mpimg
    return mpimg.imread(filename)


def createBgImage(dim):
    return np.ones((dim, dim, 3))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************



import os
import json
import datetime as dt
from collections import OrderedDict
from configparser import ConfigParser  # FIXME Does not work in Python3

import pyworkflow as pw
import pyworkflow.object as pwobj
from pyworkflow.mapper import SqliteMapper
from pyworkflow.utils import isModuleAFolder


class ProjectSettings(pwobj.OrderedObject):
    """ Store settings related to a project. """

    COLOR_MODE_STATUS = 0
    COLOR_MODE_LABELS = 1
    COLOR_MODE_AGE = 2
    COLOR_MODES = (COLOR_MODE_STATUS, COLOR_MODE_LABELS, COLOR_MODE_AGE)

    def __init__(self, confs={}, **kwargs):
        pwobj.OrderedObject.__init__(self, **kwargs)
        self.config = ProjectConfig()
        # Store the current view selected by the user
        self.currentProtocolsView = pwobj.String()
        # Store the color mode: 0= Status, 1=Labels, ...
        self.colorMode = pwobj.Integer(ProjectSettings.COLOR_MODE_STATUS)
        # Store graph nodes positions and other info
        self.nodeList = NodeConfigList()
        self.labelsList = LabelsList()  # Label list
        self.mapper = None  # This should be set when load, or write
        self.runsView = pwobj.Integer(1)  # by default the graph view
        self.readOnly = pwobj.Boolean(False)
        self.runSelection = pwobj.CsvList(int)  # Store selected runs
        self.dataSelection = pwobj.CsvList(int)  # Store selected runs
        # Some extra settings stored, now mainly used
        # from the webtools
        # Time when the project was created
        self.creationTime = pwobj.String(dt.datetime.now())
        # Number of days that this project is active
        # if None, the project will not expire
        # This is used in webtools where a limited time
        # is allowed for each project
        self.lifeTime = pwobj.Integer()
        # Set a disk quota for the project (in Gb)
        # if None, quota is unlimited
        self.diskQuota = pwobj.Integer()

    def commit(self):
        """ Commit changes made. """
        self.mapper.commit()

    def getRunsView(self):
        return self.runsView.get()

    def setRunsView(self, value):
        self.runsView.set(value)

    def getReadOnly(self):
        return self.readOnly.get()

    def setReadOnly(self, value):
        self.readOnly.set(value)

    def getCreationTime(self):
        return self.creationTime.datetime()

    def setCreationTime(self, value):
        self.creationTime.set(value)

    def getLifeTime(self):
        return self.lifeTime.get()

    def setLifeTime(self, value):
        self.lifeTime.set(value)

    def getConfig(self):
        return self.config

    def getProtocolView(self):
        return self.currentProtocolsView.get()

    def setProtocolView(self, protocolView):
        """ Set the new protocol Menu given its index.
        The new ProtocolMenu will be returned.
        """
        self.currentProtocolsView.set(protocolView)

    def getColorMode(self):
        return self.colorMode.get()

    def setColorMode(self, colorMode):
        """ Set the color mode to use when drawing the graph.
        """
        self.colorMode.set(colorMode)

    def statusColorMode(self):
        return self.getColorMode() == self.COLOR_MODE_STATUS

    def labelsColorMode(self):
        return self.getColorMode() == self.COLOR_MODE_LABELS

    def ageColorMode(self):
        return self.getColorMode() == self.COLOR_MODE_AGE

    def write(self, dbPath=None):
        self.setName('ProjectSettings')
        if dbPath is not None:
            self.mapper = SqliteMapper(dbPath, globals())
        else:
            if self.mapper is None:
                raise Exception("Can't write ProjectSettings without "
                                "mapper or dbPath")

        self.mapper.deleteAll()
        self.mapper.insert(self)
        self.mapper.commit()

    def getNodes(self):
        return self.nodeList

    def getNodeById(self, nodeId):
        return self.nodeList.getNode(nodeId)

    def addNode(self, nodeId, **kwargs):
        return self.nodeList.addNode(nodeId, **kwargs)

    def getLabels(self):
        return self.labelsList

    @classmethod
    def load(cls, dbPath):
        """ Load a ProjectSettings from dbPath. """
        classDict = dict(globals())
        classDict.update(pwobj.__dict__)
        mapper = SqliteMapper(dbPath, classDict)
        settingList = mapper.selectByClass('ProjectSettings')
        n = len(settingList)

        if n == 0:
            raise Exception("Can't load ProjectSettings from %s" % dbPath)
        elif n > 1:
            raise Exception("Only one ProjectSettings is expected in db, "
                            "found %d in %s" % (n, dbPath))

        settings = settingList[0]
        settings.mapper = mapper

        return settings


class ProjectConfig(pwobj.OrderedObject):
    """A simple base class to store ordered parameters"""
    def __init__(self, **args):
        pwobj.OrderedObject.__init__(self, **args)
        self.logo = pwobj.String('scipion_logo_small.gif')
        # Do not store this object, unless we implement some kind of
        # icon customization
        self._objDoStore = False


class MenuConfig(object):
    """Menu configuration in a tree fashion.
    Each menu can contains submenus.
    Leaf elements can contain actions"""
    def __init__(self, text=None, value=None,
                 icon=None, tag=None, **kwargs):
        """Constructor for the Menu config item.
        Arguments:
          text: text to be displayed
          value: internal value associated with the item.
          icon: display an icon with the item
          tag: put some tags to items
        **args: pass other options to base class.
        """
        self.text = pwobj.String(text)
        self.value = pwobj.String(value)
        self.icon = pwobj.String(icon)
        self.tag = pwobj.String(tag)
        self.shortCut = pwobj.String(kwargs.get('shortCut', None))
        self.childs = pwobj.List()
        self.openItem = pwobj.Boolean(kwargs.get('openItem', False))

    def addSubMenu(self, text, value=None, **args):
        subMenu = type(self)(text, value, **args)
        self.childs.append(subMenu)
        return subMenu

    def __iter__(self):
        for v in self.childs:
            yield v

    def __len__(self):
        return len(self.childs)

    def isEmpty(self):
        return len(self.childs) == 0


class ProtocolConfig(MenuConfig):
    """Store protocols configuration """
    def __init__(self, text=None, value=None, **args):
        MenuConfig.__init__(self, text, value, **args)
        if 'openItem' not in args:
            self.openItem.set(self.tag.get() != 'protocol_base')

    def addSubMenu(self, text, value=None, shortCut=None, **args):
        if 'icon' not in args:
            tag = args.get('tag', None)
            if tag == 'protocol':
                args['icon'] = 'python_file.gif'
            elif tag == 'protocol_base':
                args['icon'] = 'class_obj.gif'

        args['shortCut'] = shortCut
        return MenuConfig.addSubMenu(self, text, value, **args)


class NodeConfig(pwobj.Scalar):
    """ Store Graph node information such as x, y. """
    
    def __init__(self, nodeId=0, x=None, y=None, selected=False, expanded=True):
        pwobj.Scalar.__init__(self)
        # Special node id 0 for project node
        self._values = {'id': nodeId, 
                        'x': pwobj.Integer(x).get(0), 
                        'y': pwobj.Integer(y).get(0), 
                        'selected': selected, 
                        'expanded': expanded,
                        'labels': []}
        
    def _convertValue(self, value):
        """Value should be a str with comma separated values
        or a list.
        """
        self._values = json.loads(value)
            
    def getObjValue(self):
        self._objValue = json.dumps(self._values)
        return self._objValue
    
    def get(self):
        return self.getObjValue()
        
    def getId(self):
        return self._values['id']
    
    def setX(self, x):
        self._values['x'] = x
        
    def getX(self):
        return self._values['x']
    
    def setY(self, y):
        self._values['y'] = y
        
    def getY(self):
        return self._values['y']
    
    def setPosition(self, x, y):
        self.setX(x)
        self.setY(y)
        
    def getPosition(self):
        return self.getX(), self.getY()
        
    def setSelected(self, selected):
        self._values['selected'] = selected
        
    def isSelected(self):
        return self._values['selected']
    
    def setExpanded(self, expanded):
        self._values['expanded'] = expanded
        
    def isExpanded(self):
        return self._values['expanded']

    def setLabels(self, labels):
        self._values['labels'] = labels

    def getLabels(self):
        return self._values.get('labels', None)
    
    def __str__(self):
        return 'NodeConfig: %s' % self._values
    
    
class NodeConfigList(pwobj.List):
    """ Store all nodes information items and 
    also store a dictionary for quick access
    to nodes query.
    """
    def __init__(self):
        self._nodesDict = {}
        pwobj.List.__init__(self)
        
    def getNode(self, nodeId):
        return self._nodesDict.get(nodeId, None)
    
    def addNode(self, nodeId, **kwargs):
        node = NodeConfig(nodeId, **kwargs)
        self._nodesDict[node.getId()] = node
        self.append(node)
        return node
        
    def updateDict(self):
        self._nodesDict.clear()
        for node in self:
            self._nodesDict[node.getId()] = node
            
    def clear(self):
        pwobj.List.clear(self)
        self._nodesDict.clear()
        

class Label(pwobj.Scalar):
    """ Store Label information """

    def __init__(self, labelId=None, name='', color=None):
        pwobj.Scalar.__init__(self)
        # Special node id 0 for project node
        self._values = {'id': labelId,
                        'name': name,
                        'color': color}

    def _convertValue(self, value):
        """Value should be a str with comma separated values
        or a list.
        """
        self._values = json.loads(value)

    def getObjValue(self):
        self._objValue = json.dumps(self._values)
        return self._objValue

    def get(self):
        return self.getObjValue()

    def getId(self):
        return self._values['id']

    def getName(self):
        return self._values['name']

    def setName(self, newName):
        self._values['name'] = newName

    def setColor(self, color):
        self._values['color'] = color

    def getColor(self):
        return self._values.get('color', None)

    def __str__(self):
        return 'Label: %s' % self._values

    def __eq__(self, other):
        return self.getName() == other.getName()


class LabelsList(pwobj.List):
    """ Store all labels information"""
    def __init__(self):
        self._labelsDict = {}
        pwobj.List.__init__(self)

    def getLabel(self, name):
        return self._labelsDict.get(name, None)

    def addLabel(self, label):
        self._labelsDict[label.getName()] = label
        self.append(label)
        return label

    def updateDict(self):
        self._labelsDict.clear()
        for label in self:
            self._labelsDict[label.getName()] = label

    def deleteLabel(self, label):
        self._labelsDict.pop(label.getName())
        self.remove(label)

    def clear(self):
        pwobj.List.clear(self)
        self._labelDict.clear()


class ProtocolTreeConfig:
    """ Handler class that groups functions and constants
    related to the protocols tree configuration.
    """
    ALL_PROTOCOLS = "All"
    TAG_PROTOCOL_DISABLED = 'protocol-disabled'
    TAG_PROTOCOL = 'protocol'
    TAG_SECTION = 'section'
    TAG_PROTOCOL_GROUP = 'protocol_group'
    PLUGIN_CONFIG_PROTOCOLS = 'protocols.conf'

    @classmethod
    def getProtocolTag(cls, isInstalled):
        """ Return the proper tag depending if the protocol is installed or not.
        """
        return cls.TAG_PROTOCOL if isInstalled else cls.TAG_PROTOCOL_DISABLED

    @classmethod
    def isAFinalProtocol(cls, v, k):
        if (issubclass(v, pw.viewer.ProtocolViewer) or
                v.isBase() or v.isDisabled()):
            return False

        # To remove duplicated protocol, ProtMovieAlignment turns into OF:
        # ProtMovieAlignment = XmippProtOFAlignment
        return v.__name__ == k

    @classmethod
    def __addToTree(cls, menu, item, checkFunction=None):
        """ Helper function to recursively add items to a menu.
        Add item (a dictionary that can contain more dictionaries) to menu
        If check function is added will use it to check if the value must be added.
        """
        children = item.pop('children', [])

        if checkFunction is not None:
            add = checkFunction(item)
            if not add:
                return
        subMenu = menu.addSubMenu(**item)  # we expect item={'text': ...}
        for child in children:
            cls.__addToTree(subMenu, child, checkFunction)  # add recursively to sub-menu

        return subMenu

    @classmethod
    def __inSubMenu(cls, child, subMenu):
        """
        Return True if child belongs to subMenu
        """
        for ch in subMenu:
            if child['tag'] == cls.TAG_PROTOCOL:
                if not ch.value.empty() and ch.value == child['value']:
                    return ch
            elif ch.text == child['text']:
                return ch
        return None

    @classmethod
    def _orderSubMenu(cls, session):
        """
        Order all children of a given session:
        The protocols first, then the sessions(the 'more' session at the end)
        """
        lengthSession = len(session.childs)
        if lengthSession > 1:
            childs = session.childs
            lastChildPos = lengthSession - 1
            if childs[lastChildPos].tag == cls.TAG_PROTOCOL:
                for i in range(lastChildPos - 1, -1, -1):
                    if childs[i].tag == cls.TAG_PROTOCOL:
                        break
                    else:
                        tmp = childs[i+1]
                        childs[i+1] = childs[i]
                        childs[i] = tmp
            else:
                for i in range(lastChildPos - 1, -1, -1):
                    if childs[i].tag == cls.TAG_PROTOCOL:
                        break
                    elif 'more' in str(childs[i].text).lower():
                        tmp = childs[i+1]
                        childs[i+1] = childs[i]
                        childs[i] = tmp

    @classmethod
    def __findTreeLocation(cls, subMenu, children, parent):
        """
        Locate the protocol position in the given view
        """
        for child in children:
            sm = cls.__inSubMenu(child, subMenu)
            if sm is None:
                cls.__addToTree(parent, child, cls.__checkItem)
                cls._orderSubMenu(parent)
            elif child['tag'] == cls.TAG_PROTOCOL_GROUP or child['tag'] == cls.TAG_SECTION:
                cls.__findTreeLocation(sm.childs, child['children'], sm)

    @classmethod
    def __checkItem(cls, item):
        """ Function to check if the protocol has to be added or not.
        Params:
            item: {"tag": "protocol", "value": "ProtImportMovies",
                   "text": "import movies"}
        """
        if item["tag"] != cls.TAG_PROTOCOL:
            return True

        # It is a protocol as this point, get the class name and
        # check if it is disabled
        protClassName = item["value"]
        protClass = pw.Config.getDomain().getProtocols().get(protClassName)

        return False if protClass is None else not protClass.isDisabled()

    @classmethod
    def __addAllProtocols(cls, domain, protocols):
        # Add all protocols
        # FIXME: Check why this import is here
        allProts = domain.getProtocols()

        # Sort the dictionary
        allProtsSorted = OrderedDict(sorted(allProts.items(),
                                            key=lambda e: e[1].getClassLabel()))
        allProtMenu = ProtocolConfig(cls.ALL_PROTOCOLS)
        packages = {}

        # Group protocols by package name
        for k, v in allProtsSorted.items():
            if cls.isAFinalProtocol(v, k):
                packageName = v.getClassPackageName()
                # Get the package submenu
                packageMenu = packages.get(packageName)

                # If no package menu available
                if packageMenu is None:
                    # Add it to the menu ...
                    packageLine = {"tag": "package", "value": packageName,
                                   "text": packageName}
                    packageMenu = cls.__addToTree(allProtMenu, packageLine)

                    # Store it in the dict
                    packages[packageName] = packageMenu

                # Add the protocol
                tag = cls.getProtocolTag(v.isInstalled())

                protLine = {"tag": tag, "value": k,
                            "text": v.getClassLabel(prependPackageName=False)}

                # If it's a new protocol
                if v.isNew() and v.isInstalled():
                    # add the new icon
                    protLine["icon"] = "newProt.gif"

                cls.__addToTree(packageMenu, protLine)

        protocols[cls.ALL_PROTOCOLS] = allProtMenu

    @classmethod
    def __addProtocolsFromConf(cls, protocols, protocolsConfPath):
        """
        Load the protocols in the tree from a given protocols.conf file,
        either the global one in Scipion or defined in a plugin.
        """
        # Populate the protocols menu from the plugin config file.
        if os.path.exists(protocolsConfPath):
            cp = ConfigParser()
            cp.optionxform = str  # keep case
            cp.read(protocolsConfPath)
            #  Ensure that the protocols section exists
            if cp.has_section('PROTOCOLS'):
                for menuName in cp.options('PROTOCOLS'):
                    if menuName not in protocols:  # The view has not been inserted
                        menu = ProtocolConfig(menuName)
                        children = json.loads(cp.get('PROTOCOLS', menuName))
                        for child in children:
                            cls.__addToTree(menu, child, cls.__checkItem)
                        protocols[menuName] = menu
                    else:  # The view has been inserted
                        menu = protocols.get(menuName)
                        children = json.loads(cp.get('PROTOCOLS',
                                                     menuName))
                        cls.__findTreeLocation(menu.childs, children, menu)

    @classmethod
    def load(cls, domain, protocolsConf):
        """ Read the protocol configuration from a .conf file similar to the
        one in scipion/config/protocols.conf,
        which is the default one when no file is passed.
        """
        protocols = OrderedDict()
        # Read the protocols.conf from Scipion (base) and create an initial
        # tree view
        cls.__addProtocolsFromConf(protocols, protocolsConf)

        # Read the protocols.conf of any installed plugin
        pluginDict = domain.getPlugins()
        pluginList = pluginDict.keys()
        for pluginName in pluginList:
            try:

                # if the plugin has a path
                if isModuleAFolder(pluginName):
                    # Locate the plugin protocols.conf file
                    protocolsConfPath = os.path.join(
                                            pluginDict[pluginName].__path__[0],
                                            cls.PLUGIN_CONFIG_PROTOCOLS)
                    cls.__addProtocolsFromConf(protocols, protocolsConfPath)

            except Exception as e:
                print('Failed to read settings. The reported error was:\n  %s\n'
                      'To solve it, fix %s and run again.' % (
                            e, os.path.abspath(protocolsConfPath)))

            # Add all protocols to All view
        cls.__addAllProtocols(pw.Config.getDomain(), protocols)

        return protocols



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
MPI utilities. runJobMPI and runJobMPISlave send and receive the commands
to execute, in the given directory and with the given environment.
"""


import os
from time import time, sleep
from pickle import dumps, loads

from pyworkflow import Config

from .process import buildRunCommand, runCommand

from pyworkflow.utils.utils import getLocalHostName

TIMEOUT = 60  # seconds trying to send/receive data through a socket

TAG_RUN_JOB = 1000


def send(command, comm, dest, tag):
    """ Send command in a non-blocking way and raise exception on error. """

    # This function blocks, but it uses the isend() function (which is
    # nonblocking) and sleeps without using the cpu while we try to send.
    # Also, if we cannot send after TIMEOUT seconds, raise exception.

    if command.startswith('env='):
        print("Sending environment to %d" % dest)
    else:
        print("Sending command to %d: %s" % (dest, command))

    # Send command with isend()
    req_send = comm.isend(command, dest=dest, tag=tag)
    t0 = time()
    while not req_send.test()[0]:
        sleep(1)
        if time() - t0 > TIMEOUT:
            raise Exception("Timeout in process %d, cannot send command "
                            "to slave." % os.getpid())

    # Receive the result in a non-blocking way too (with irecv())
    req_recv = comm.irecv(source=dest, tag=tag)
    while True:
        done, result = req_recv.test()
        if done:
            break
        sleep(1)

    if result != 0:  # result will then be a string with the error
        raise Exception(str(result))


def runJobMPI(programname, params, mpiComm, mpiDest,
              numberOfMpi=1, hostConfig=None,
              env=None, cwd=None, gpuList=None):
    """ Send the command to the MPI node in which it will be executed. """

    command = buildRunCommand(programname, params, numberOfMpi, hostConfig,
                              env, gpuList=gpuList)
    if cwd is not None:
        send("cwd=%s" % cwd, mpiComm, mpiDest, TAG_RUN_JOB+mpiDest)
    if env is not None:
        send("env=%s" % dumps(env), mpiComm, mpiDest, TAG_RUN_JOB+mpiDest)

    send(command, mpiComm, mpiDest, TAG_RUN_JOB+mpiDest)


def runJobMPISlave(mpiComm):
    """ This slave will be receiving commands to execute
    until 'None' is received. 
    """
    rank = mpiComm.Get_rank()
    hostname = getLocalHostName()
    print("Running runJobMPISlave: ", rank)

    exitResult = 0
    msg = "Timeout in process %d, cannot send result to master."

    # Listen for commands until we get 'None'
    cwd = None  # We run without changing directory by default
    env = None  # And we don't change the environment either!
    
    while True:
        # Receive command in a non-blocking way
        req_recv = mpiComm.irecv(source=0, tag=TAG_RUN_JOB+rank)
        while True:
            done, command = req_recv.test()
            if done:
                break
            sleep(1)

        print("Slave %s(rank %d) received command." % (hostname, rank))
        if command == 'None':
            print("  Stopping...")
            return

        # Run the command and get the result (exit code or exception)
        try:
            if command.startswith("cwd="):
                cwd = command.split("=", 1)[-1]
                print("  Changing to dir %s ..." % cwd)
            elif command.startswith("env="):
                env = loads(command.split("=", 1)[-1])
                print("  Setting the environment...")
                if Config.debugOn():
                    print(env)
            else:
                print("  %s" % command)
                runCommand(command, cwd=cwd, env=env)
                cwd = None  # unset directory
                env = None  # unset environment
        except Exception as e:
            msg = "Timeout in process %d, cannot send error message to master."
            exitResult = str(e)

        # Communicate to master, either error os success
        req_send = mpiComm.isend(exitResult, dest=0, tag=TAG_RUN_JOB+rank)
        t0 = time()
        while not req_send.test()[0]:
            sleep(1)
            if time() - t0 > TIMEOUT:
                print(msg % os.getpid())
                break
